// code by jph

matrix<float> wCam2gnd[2];
// matrix contain ground coordinate corresponding to pixel pin in cam image
matrix<float> jCam2gnd(int cam,int pin) {
  float z0=pin%320-160; z0/=100.f;
  float z1=pin/320-120; z1/=100.f;
  matrix<float> piz(5,1);
    piz.data[0]=z0*z1;
    piz.data[1]=z1*z1;
    piz.data[2]=z0;
    piz.data[3]=z1;
    piz.data[4]=1;
  return jCam*wCam2gnd[cam]*piz;
}

matrix<float> wGnd2cam[2];
// returns number of pixel of gnd-coordinate in cam image
// returns -1     if coordinate is not within the cam image
int jGnd2cam(int cam,const matrix<float> &pos) {
  matrix<float> tmp=jCin*pos;
  matrix<float> piz(5,1);
    piz.data[0]=tmp.data[0]*tmp.data[1];
    piz.data[1]=tmp.data[1]*tmp.data[1];
    piz.data[2]=tmp.data[0];
    piz.data[3]=tmp.data[1];
    piz.data[4]=1;
  matrix<float> ret=wGnd2cam[cam]*piz;
  int pix=(int)(ret.data[0]*100+160);
  int piy=(int)(ret.data[1]*100+120);
  if (0<pix&&pix<320&&0<piy&&piy<240)
    return piy*320+pix;
  else
    return -1;
}

int jSingleton(int msk) {
  int c1;
  for (c1=0;c1<8;c1++) {
    if (msk&1)
      return 1<<c1;
    msk>>=1;
  }
}

// translate laser.distanceArray to local coordinates 
// camera <-> ground conversion matrices (left and right)
// compute field of view of camera[0] and camera[1] cut off by jLaser
// returns false if laser scan has problems
bool jSensors(void) {
  jDEBUG("jSensors")
  int c1,c2;
  // update internals  
  // BEGIN: camera and buffer
  if (jBuffer)
    SDL_FreeSurface(jBuffer);
  jBuffer=NULL;
  if (camera)
    jBuffer=camera;
  camera=NULL;
  jDEBUG("get cam")
  if (jTic[jTOC_NEXT_PTU]<jTic[jTOC_LAST_SENSORS]&&
      fabs(jVel.data[0]) <jSPEED_BLUR) {
    TBitmap myTBitmap;
    SDL_RWops *mySDL_RWops;
    SDL_Surface *mySDL_Surface;
    if (ctrl.IC_GetImage(jCAMERA_MONO,&myTBitmap)) {
      if (mySDL_RWops=SDL_RWFromMem(myTBitmap.data,myTBitmap.data_size)) {
        if (mySDL_Surface=IMG_LoadJPG_RW(mySDL_RWops)) {
          camera=SDL_ConvertSurface(mySDL_Surface,&jSDL_PixelFormat,0);
          SDL_FreeSurface(mySDL_Surface);
        }
        SDL_FreeRW(mySDL_RWops);
      }
      myTBitmap.free_data();
    }
  }
  jDEBUG("got cam")
  // END: camera and buffer
  // BEGIN: laser
  c1=0; ctrl.ASRobo_GetBumperScan(bumper,&c1);
  c1=0; ctrl.ASRobo_GetLaserScan(laser,&c1);
  if (laser.distanceArray.size()!=jLASER)
    return false;  
  jTic[jTOC_LAST_SENSORS]=jCLOCK_FUNC;  
  jMover->visual=jTic[jTOC_LAST_SENSORS];
  float dst; // the maximum distance ever measured is 81.910
  float dst_min=100.0f;
  float dst_max=000.0f;
  double ang_abs=M_PI/2.;
  for (c1=0;c1<jLASER;c1++) {
    dst=laser.distanceArray[c1].distance;
    jLaserLocal.data[c1*3+0] =jLaser.data[c1*2+0]=sin(ang_abs)*dst;
    jLaserLocal.data[c1*3+1] =jLaser.data[c1*2+1]=cos(ang_abs)*dst;
    jLaserLocal.data[c1*3+1]+=jLASER_OFFSET;
    jLaserLocal.data[c1*3+2] =1.0f;
    ang_abs-=jDEG2RAD;
    dst_min =MIN(dst_min,dst);
    dst_max =MAX(dst_max,dst);
  }
  // END: laser
  // BEGIN: bumper
  jBumperOld=jBumper;
  jBumper=0;
  jBumperCount=0;
  for (c1=0;c1<bumper.distanceArray.size();c1++)
    if (0<bumper.distanceArray[c1].distance)
      jBumper|=1<<c1;
  if (jBumper) {
    jBumperLast=jBumper;
    jTic[jTOC_LAST_BUMPER]=jTic[jTOC_LAST_SENSORS];
  }
  if (jTog[jTOG_BUMPER_COUNTING]) {
    if (((bool)jBumper^(bool)jBumperOld)&&jTic[jTOC_LAST_SENSORS]<=jTic[jTOC_LAST_BUMPER]+1200) // &1 means only front bumper
      jInt[jINT_BUMPER_COUNT]++;
    if (jTic[jTOC_LAST_BUMPER]+1200<jTic[jTOC_LAST_SENSORS]) {
      jBumperCount=jInt[jINT_BUMPER_COUNT];
      jInt[jINT_BUMPER_COUNT]=0;
    }
  } else
    jInt[jINT_BUMPER_COUNT]=0;
  if (jBumper) {
    sprintf(jChar,"bumper pressed %i",jBumper);
    jBulletin.append(new std::string(jChar));
  }
  //if (jBumperCount) {
  //  sprintf(jChar,"bumper counted %i at %i",jBumperCount,jBumperLast);
  //  jDeclared.prepend(new std::string(jChar));
  //}
  // END: bumper
  // BEGIN: camera <-> ground conversion matrices (left and right)
  float fp=0;
  float ft=jOldPTU.data[1];
  matrix<float> fpt(6,1);
    fpt.data[0]=fp*fp;
    fpt.data[1]=ft*fp;
    fpt.data[2]=ft*ft;
    fpt.data[3]=fp;
    fpt.data[4]=ft;
    fpt.data[5]=1;
  matrix<float> tmp;
  tmp=jCamL2gnd*fpt; // left camera to ground
  for (c1=0;c1<5;c1++) {
    wCam2gnd[0].data[c1*3+0]=tmp.data[c1+0];
    wCam2gnd[0].data[c1*3+1]=tmp.data[c1+5];
  }
  tmp=jCamR2gnd*fpt; // right camera to ground
  for (c1=0;c1<5;c1++) {
    wCam2gnd[1].data[c1*3+0]=tmp.data[c1+0];
    wCam2gnd[1].data[c1*3+1]=tmp.data[c1+5];
  }
  tmp=jGnd2camL*fpt; // ground to left camera
  for (c1=0;c1<5;c1++) {
    wGnd2cam[0].data[c1*2+0]=tmp.data[c1+0];
    wGnd2cam[0].data[c1*2+1]=tmp.data[c1+5];
  }
  tmp=jGnd2camR*fpt; // ground to right camera
  for (c1=0;c1<5;c1++) {
    wGnd2cam[1].data[c1*2+0]=tmp.data[c1+0];
    wGnd2cam[1].data[c1*2+1]=tmp.data[c1+5];
  }    
  float ang=jOldPTU.data[0];  
  jCam.data[0]= cos(ang);
  jCam.data[1]=-sin(ang);
  jCam.data[3]=-jCam.data[1];
  jCam.data[4]= jCam.data[0];
  jCam.data[6]=(cos(ang)-1)*jPTUPAN_RAD;  
  jCam.data[7]=   -sin(ang)*jPTUPAN_RAD+jPTU_OFFSET;
  jCin=jCam.solve(matrix<float>(3,3,"eye"));
  // compute field of view of camera[0] and camera[1] cut off by jLaser
  matrix<float> pos;
  int deg;
  for (c1=0;c1<2;c1++) { // both cameras
    for (c2=0;c2<jFOV;c2++) {
      pos=jCam2gnd(c1,jFoV[c2]);
      jSight[c1].data[c2*3+0]=pos.data[0];
      jSight[c1].data[c2*3+1]=pos.data[1];
      pos.data[1]-=.1f; // move 10 cm closer
      pos.data[2]=0;
      deg=(int)(atan2(-pos.data[0],pos.data[1])*57.295779f+90);
      if (0<deg&&deg<180)
        if (laser.distanceArray[deg].distance<pos.norm(2)) {
          jSight[c1].data[c2*3+0]=jLaser.data[deg*2+0];
          jSight[c1].data[c2*3+1]=jLaser.data[deg*2+1]+jLASER_OFFSET;
        }
    }
  }
  // END: camera <-> ground conversion matrices (left and right)
  // BEGIN: extract RGB values from camera image
  if (camera) {
    unsigned int col;
    SDL_LockSurface(camera);
    Uint32* cam=(Uint32*)(camera->pixels);     // to manipulate pixel array  
    c2=0;
    for (c1=0;c1<jCAMERA;c1++) {
      col=cam[c1];
      jCameraRGB[c2++]=(col&0xff0000)>>16;
      jCameraRGB[c2++]=(col&0xff00  )>> 8;
      jCameraRGB[c2++]= col&0xff;
    }
    SDL_UnlockSurface(camera);
  }
  // END: extract RGB values from camera image  
  
  if (jJoystick) {    
    jAxis.data[0]=-SDL_JoystickGetAxis(jJoystick,1)*jJOYSTICK_FACTOR;
    jAxis.data[0]*=jJOYSTICK_DEADZONE<fabs(jAxis.data[0]);    
    jAxis.data[1]= SDL_JoystickGetAxis(jJoystick,0)*jJOYSTICK_FACTOR;
    jAxis.data[1]*=jJOYSTICK_DEADZONE<fabs(jAxis.data[1]);
    jAxis.data[2]=-SDL_JoystickGetAxis(jJoystick,3)*jJOYSTICK_FACTOR;
    jAxis.data[2]*=jJOYSTICK_DEADZONE<fabs(jAxis.data[2]);
    jAxis.data[3]=-SDL_JoystickGetAxis(jJoystick,2)*jJOYSTICK_FACTOR;
    jAxis.data[3]*=jJOYSTICK_DEADZONE<fabs(jAxis.data[3]);
    
    for (c1=0;c1<jButton.numel;c1++)
      jButton.data[c1]=SDL_JoystickGetButton(jJoystick,c1);
  }
  return true;
}

// performs histogram equalization into jGoggle
// to enhance user interface
void jGoggles(void) {
  if (!camera) return;
  int c1,c2;
  int sum,exb;
  unsigned int col;
  int jCameraHist[257]; // 256+1 for safety, 256 = nirvana
  for (c1=0;c1<257;c1++)
    jCameraHist[c1]=0;
  // BEGIN: compile histogram
  for (c1=0;c1<jCAMERA3;c1++)
    jCameraHist[jCameraRGB[c1]]++;
  sum=1; // this could be good for safety
  exb=0;  
  for (c1=0;c1<256;c1++) {
    sum+=jCameraHist[c1];  
    exb+=jCameraHist[c1]*c1;
    jCameraHist[c1]=sum<<8;
  }
  jCameraBrightness=exb/(float)sum;
  for (c1=0;c1<256;c1++) {
    jCameraHist[c1]/=sum;
    jCameraHist[c1] =(jCameraHist[c1]+c1)/2;
  }
  // END: compile histogram
  // BEGIN: inject to night vision goggles image
  SDL_LockSurface(jGoggle);
  Uint32* gog=(Uint32*)(jGoggle->pixels);     // to manipulate pixel array    
  c2=0;
  for (c1=0;c1<jCAMERA;c1++) {
    col=0xff000000;
    col|=jCameraHist[jCameraRGB[c2++]]<<16;
    col|=jCameraHist[jCameraRGB[c2++]]<<8;
    col|=jCameraHist[jCameraRGB[c2++]];
    gog[c1]=col;
  }
  SDL_UnlockSurface(jGoggle);
  // END: inject to night vision goggles image
}

void jMarkerSet(int pix,int piy) {
  matrix<float> nrm(3,1);
  int pin;
  pin=(pix+piy*jCAMERA_W)*3;
  nrm.data[0]=jCameraRGB[pin+0];
  nrm.data[1]=jCameraRGB[pin+1];
  nrm.data[2]=jCameraRGB[pin+2];
  jMarker.data[12]=nrm.normalize(2);
  jMarker.data[ 0]=nrm.data[0];
  jMarker.data[ 1]=nrm.data[1];
  jMarker.data[ 2]=nrm.data[2];
  jMarker.glOrtho();
}

// resampling of the laser points (number of samples varies, samples are 4 cm apart)
void jLasers(void) {
  jDEBUG("jLasers")
  int c1;
  int ind,tot=0;
  int end,cnt,num,pos;
  float del,lam;
  float dif[181];
  float acc[181];
  matrix<float> tok;
  matrix<float> tmp(2,1);
  while (jLaserList.token(tok))
    jLaserList.remove();
  for (c1=0;c1<jLaser.size[1]-1;c1++) {
    tmp.data[0]=jLaser.data[c1*2+0]-jLaser.data[c1*2+2];
    tmp.data[1]=jLaser.data[c1*2+1]-jLaser.data[c1*2+3];
    dif[c1]=tmp.norm(2);
  }
  ind=0;
  dif[180]=1000;     // high value
  acc[0  ]=   0;
  while (ind<jLaser.size[1]) {
    end=ind;
    while (dif[end]<(laser.distanceArray[end].distance+laser.distanceArray[end+1].distance)*jLASER_TANGENS&&laser.distanceArray[end  ].distance<jLASER_IGNORE&&laser.distanceArray[end+1].distance<jLASER_IGNORE) {
      acc[end-ind+1]=acc[end-ind]+dif[end];
      end++;
    }
    cnt=end-ind;
    if (cnt&&jLASER_SAMPLING<acc[cnt]) {
      num=MAX(1,(int)(acc[cnt]/jLASER_SAMPLING));
      del=acc[cnt]/(float)num;
      matrix<float> seq(2,num+1);
      seq.data[0]=jLaser.data[ind*2+0];
      seq.data[1]=jLaser.data[ind*2+1];
      pos=0;
      for (c1=1;c1<num;c1++) {
        while (acc[pos+1]<del*c1)
          pos++;
        lam=(del*c1-acc[pos])/(acc[pos+1]-acc[pos]);
        seq.data[c1*2+0]=lam*jLaser.data[(ind+pos+1)*2+0]+(1-lam)*jLaser.data[(ind+pos)*2+0];
        seq.data[c1*2+1]=lam*jLaser.data[(ind+pos+1)*2+1]+(1-lam)*jLaser.data[(ind+pos)*2+1];
      }
      seq.data[seq.numel-2+0]=jLaser.data[end*2+0];
      seq.data[seq.numel-2+1]=jLaser.data[end*2+1];
      tot+=seq.size[1];
      if (tot<jLASER_SAMPLES)
        jLaserList.append(seq);      
    } else {
      matrix<float> seq(2,1);
      seq.data[0]=jLaser.data[ind*2+0];
      seq.data[1]=jLaser.data[ind*2+1];      
      tot+=seq.size[1];
      if (tot<jLASER_SAMPLES)
        jLaserList.append(seq);      
    }
    ind=end+1;
  }
  // tot has the number of (new) samples
}

// jSpeaks initializes one message (idea) with a collection of texts
void jSpeaks(int msg,int typ,int period,int flash) {
  int c1=20;
  jSpeak[msg].typ=typ;
  jSpeak[msg].stamp=-10000;
  jSpeak[msg].period=period;
  jSpeak[msg].flash=flash;
  jSpeak[msg].text=array<std::string*>(1,&c1,NULL);
  jSpeak[msg].text.numel=0;
  jSpeak[msg].appendix=NULL;
}

// jSpeaks adds to one message (idea) one text
void jSpeaks(int msg,char* txt) {
  jSpeak[msg].text.data[jSpeak[msg].text.numel++]=new std::string(txt);
}

// jSpeaks adds to one message (idea) one text
void jSpeaks(char* txt) {
  std::string str(txt);
	cgim.send(std::string("J2B2_TTS2"),str.c_str(),str.size(),GIMI_TTS_SERVICE_ID,0);
	//delete str;
}

// jSpeaks handles message
void jSpeaks(int msg) {
  std::string* str;
  if (jSpeakTypes&jSpeak[msg].typ)
    if (jSpeak[msg].stamp+jSpeak[msg].period<jTic[jTOC_LAST_SENSORS]) {
      str=jSpeak[msg].text.data[(int)(jSpeak[msg].text.numel*jRandu.token())];
      cgim.send(std::string("J2B2_TTS2"),str->c_str(),str->size(),GIMI_TTS_SERVICE_ID,0);
      //jDeclared.prepend(new std::string(*str)); // TODO no display during development
      str=jSpeak[msg].appendix;
      if (str)
        cgim.send(std::string("J2B2_TTS2"),str->c_str(),str->size(),GIMI_TTS_SERVICE_ID,0);
      jSpeak[msg].stamp=jTic[jTOC_LAST_SENSORS];
      if (jSpeak[msg].typ&jSPEAK_TYPE_PROBLEM)
        jFlashingLaser.size[1]=MAX(jFlashingLaser.size[1],2*jSpeak[msg].flash-1);
      if (jSpeak[msg].typ&jSPEAK_TYPE_EXPLAIN)
        jFlashingLight.size[1]=MAX(jFlashingLight.size[1],2*jSpeak[msg].flash-1);
    }
}

// evaluates the laser scan to determine if robot is too close to obstacle
// function resets wHazard and jHazard
void jHazards(void) {
  jDEBUG("hazards")
  int c1,cnt=0;
  float iR0=1.0f/jROBOT_RADIUS;
  float iR1=0.5f/jROBOT_RADIUS;
  float a0,a1;
  float dst;
  jHazard=jLASER_IGNORE;
  for (c1=0;c1<jLASER;c1++) {
    a0=jLaser.data[c1*2+0]*iR0;
    a1=jLaser.data[c1*2+1]*iR1;
    dst=sqrt(a0*a0+a1*a1);
    if (dst<jHazard)
      jHazard=dst;
  }
  jHazard-=1.0f;
}

// creates polygon star-convex with resp to laser 
// returns coordinates with respect to robot
// specified angular resolution
matrix<float> jExplode(int res,float cut,float ofs) {
  jDEBUG("jExplode")
  int c1,c2,c3;
  matrix<float> tmp(2,1);
  matrix<float> rot(2,2);
  matrix<float> pnt(2,jLASER_SAMPLES);    
  matrix<float> ret(3,res+1);                             // local coordinates
  for (c1=0;c1<=res;c1++) {
    ret.data[c1*3+1]=jLASER_OFFSET;
    ret.data[c1*3+2]=1.0f;
  }
  pnt.numel=0;
  matrix<float> seq;
  while (jLaserList.token(seq)) {                         // add resampled laser points if ...
    for (c1=0;c1<seq.size[1];c1++) {
      tmp.data[0]=seq.data[c1*2+0];
      tmp.data[1]=seq.data[c1*2+1];
      if (tmp.norm(2)<cut+ofs&&pnt.numel<pnt.size[1]<<1) {
        pnt.data[pnt.numel+0]=tmp.data[0];
        pnt.data[pnt.numel+1]=tmp.data[1];
        pnt.numel+=2;
      }
    }
  }
  pnt.size[1]=pnt.numel/2;
  if (!pnt.numel)
    return ret;
  float len,dst,ins;
  float of2=ofs*ofs;
  double ang_abs=M_PI/2.0;
  double ang_inc=M_PI/(double)res;    
  matrix<float> trf;
  for (c1=0;c1<=res;c1++) {
    rot.data[0]= cos(ang_abs);
    rot.data[1]= sin(ang_abs);
    rot.data[2]=-rot.data[1];
    rot.data[3]= rot.data[0];
    trf=rot*pnt;
    len=cut;
    for (c2=0;c2<trf.numel;c2+=2) {
      ins=trf.data[c2];
      if (-ofs<ins&&ins<ofs&&-jPERMILLE<trf.data[c2+1]) { // point is in corridor
        dst=trf.data[c2+1]-sqrt(fabs(of2-ins*ins));
        if (dst<len)                                      // reduce length of corridor
          len=dst;
      }
    }
    if (len<0)
      len=0;
    ret.data[c1*3+0]+=rot.data[1]*len;
    ret.data[c1*3+1]+=rot.data[0]*len;
    ang_abs-=ang_inc;
  }  
  return ret;
}

// writes surface of palette
void jPaletteUpdate(void) {
  int c1,cnt=0,lim=jPALETTE_W*jPALETTE_H;
  bucket<unsigned int,unsigned int> buc;
  SDL_FillRect(jPaletteDraw,NULL,0);
  SDL_LockSurface(jPaletteDraw);
  Uint32* pal=(Uint32*)(jPaletteDraw->pixels);      
  for (c1=0;c1<jPALETTE;c1++)
	  while (jPalette.table->data[c1]->token(buc))
	    if (cnt<lim)
	      pal[cnt++]=buc.data|0xff000000;    
  SDL_UnlockSurface(jPaletteDraw);
}

// adds colors to jPalette
void jPixels(void) {
  if (!camera) return;
  int c1,c2;
  int RED,GRE,BLU;
  float prj0,prj1,prj2,rad=jMarker.data[13]*jMarker.data[13];
  int cnt=0;                    // number of yellow pixels
  unsigned int nul;
  SDL_LockSurface( camera);     // to manipulate pixel array
  SDL_LockSurface(jCamera);     // to manipulate pixel array
  Uint32* myPixels=(Uint32*)( camera->pixels);
  Uint32* myCamera=(Uint32*)(jCamera->pixels);
  c2=jCAMERA_W*3;
  for (c1=jCAMERA_W;c1<jCAMERA;c1++) {        // loop over all pixels and detect yellow    
    RED=jCameraRGB[c2++];
    GRE=jCameraRGB[c2++];
    BLU=jCameraRGB[c2++];
    prj0=jMarker.data[ 0]*RED+jMarker.data[ 1]*GRE+jMarker.data[ 2]*BLU-jMarker.data[12];
    prj1=jMarker.data[ 4]*RED+jMarker.data[ 5]*GRE+jMarker.data[ 6]*BLU;
    prj2=jMarker.data[ 8]*RED+jMarker.data[ 9]*GRE+jMarker.data[10]*BLU;
    if (prj0*prj0+jMarker.data[14]*(prj1*prj1+prj2*prj2)<rad) 
      jPalette.store(myPixels[c1],myPixels[c1]);
    if (jPalette.find(myPixels[c1],nul))
      myCamera[c1]=0xffff00ff;    
  }  
  SDL_UnlockSurface(jCamera);
  SDL_UnlockSurface( camera);
  jPaletteUpdate();
}


// looks for jPalette in camera 
int jPixels(int* pins) {
  if (!camera) return 0;
  int c1,cnt=0;
  int RED,GRE,BLU;
  unsigned int nul;
  SDL_LockSurface( camera);
  SDL_LockSurface(jCamera);
  Uint32* myPixels=(Uint32*)( camera->pixels);
  Uint32* myCamera=(Uint32*)(jCamera->pixels);
  for (c1=0;c1<jCAMERA;c1++)                  // REVISE!!!
    jCameraMask[c1]=0;
  for (c1=jCAMERA_W;c1<jCAMERA;c1++)          // loop over all pixels and detect yellow
    if (jPalette.find(myPixels[c1],nul)) {
      cnt++;
      jCameraMask[c1]=cnt;                    // store address instead of color! to identifz stones            
      myCamera[c1]=myPixels[c1]^0x00ffffff;   // invert colors
      jCameraTree[cnt]=new sutree<int>(c1);   // MAKESET, node has pixel address as content      
      if (jCameraMask[c1-1])                  // check previous pixel
        jCameraTree[cnt]->join(jCameraTree[jCameraMask[c1-1]]);    // UNION
      if (jCameraMask[c1-320])                // check pixel one row above
        jCameraTree[cnt]->join(jCameraTree[jCameraMask[c1-320]]);  // UNION
    }
  SDL_UnlockSurface( camera);
  SDL_UnlockSurface(jCamera);
  int pin,ref=0,num=0;                        // number of stones
  for (c1=1;c1<=cnt;c1++) {                   // loop over all yellow pixels / nodes of union-data struct collecting unique representatives
    if (pin=jCameraTree[c1]->fset()->data) {
      pins[num++]=pin;
      jCameraTree[c1]->fset()->data=0;
      if (num==jPINS)
        break;
    }
  }
  for (c1=1;c1<=cnt;c1++)
    delete jCameraTree[c1];                   // delete disjoint set union data struct
  return num;  
}

void jVisuals(void) {
  jDEBUG("jVisuals")
  int c1,c2,c3;
  int pin,deg;
  int pins[jPINS];
  jInt[jINT_MARKER_PIN]=0;  
  jStone.size[1]=jStone.numel=0;
  // BEGIN: commented out only to enhance overall performance
  /*
  if (!jTog[jTOG_MARKER_CALIBRATION]) {
    if (jPTUTIL_PROFILE_MIN<jOldPTU.data[1]) {
      jDrawString(jCamera,jFond,"profile",jCOL1,0);
      pin=jPixels(pins);
      if (pin) {
        matrix<float> pos;
        jInt[jINT_MARKER_PIN]=pins[0];
        for (c1=0;c1<pin;c1++) {
          pos=jCam2gnd(jCAMERA_MONO,pins[c1]);
          jStone.data[c1*3+0]=pos.data[0];
          jStone.data[c1*3+1]=pos.data[1];
          jStone.data[c1*3+2]=1;
        }
        jStone.size[1]=pin;
        jStone.numel=pin*3;          
      }
    }
    if (jPTUTIL_GESTURE_MIN<jOldPTU.data[1]&&jOldPTU.data[1]<jPTUTIL_GESTURE_MAX) {
      jDrawString(jCamera,jFond,"gesture",jCOL1,0);
      pin=jPixels(pins);
      if (pin) {
        jTic[jTOC_STOP_GESTURE]=jTic[jTOC_LAST_SENSORS]+jGESTURE_INTERRUPT;
        jInt[jINT_MARKER_PIN]=pins[0];
        jGesture.append(jInt[jINT_MARKER_PIN]);
      }
    }
  } else
    jPixels();
  */
  // END: commented out only to enhance overall performance
}

// resamples gesture according to arc length
matrix<float> jGestures(const matrix<float> &gsl,const matrix<float> &gsp) {
  int c1,ind=0;
  float lam;
  matrix<float> ret(20,1);
  ret.data[ 0]=gsp.data[ 0];
  ret.data[19]=gsp.data[19];
  for (c1=1;c1<19;c1++) {
    while (gsl.data[ind+1]<c1)
      ind++;
    lam=(c1-gsl.data[ind])/(gsl.data[ind+1]-gsl.data[ind]);
    ret.data[c1]=lam*gsp.data[ind+1]+(1-lam)*gsp.data[ind];
  }
  return ret;
}

// analyses and deletes gesture
bool jGestures(void) {
  int c1,pin=jGESTURE,cnt=0;
  int inx,iny;
  float ivx,ivy,avx,avy;
  matrix<float> gsx(jGESTURE,1);
  matrix<float> gsy(jGESTURE,1);
  ivx=ivy= 500;
  avx=avy=-500;
  while (jGesture.token(pin)&&cnt<jGESTURE) {
    gsx.data[cnt]=160-pin%jCAMERA_W;
    ivx=MIN(ivx,gsx.data[cnt]);
    avx=MAX(avx,gsx.data[cnt]);
    gsy.data[cnt]=120-pin/jCAMERA_W;
    ivy=MIN(ivy,gsy.data[cnt]);
    avy=MAX(avy,gsy.data[cnt]);
    cnt++;
  }
  jGesture.next[2]=&jGesture;           // reset tokenizer
  while (jGesture.token(pin))           // empty list
    jGesture.remove();
  if (cnt<jGESTURE_REQUIRED||avx-ivx<jGESTURE_RANGEMIN||avy-ivy<jGESTURE_RANGEMIN)
    return false;
  gsx.size[0]=gsx.numel=cnt;
  gsy.size[0]=gsy.numel=cnt;
  gsx-=(avx+ivx)*0.5f;
  gsy-=(avy+ivy)*0.5f;
  gsx*=1.0f/(avx-ivx);
  gsy*=1.0f/(avy-ivy);
  matrix<float> tmp(2,1);
  matrix<float> gsl(cnt,1);
  for (c1=1;c1<cnt;c1++) {
    tmp.data[ 0]=gsx.data[c1-1]-gsx.data[c1];
    tmp.data[ 1]=gsy.data[c1-1]-gsy.data[c1];
    gsl.data[c1]=gsl.data[c1-1]+tmp.norm(2);
  }
  gsl*=19.0f/gsl.data[cnt-1];
  jGestureResX=jGestureX*jGestures(gsl,gsx);
  jGestureResY=jGestureY*jGestures(gsl,gsy);
  avx=avy=-100;
  inx=iny=-1;
  for (c1=0;c1<jGestureResX.numel;c1++) {
    if (avx<jGestureResX.data[c1]) {
      avx=jGestureResX.data[c1];
      inx=c1;
    }
    if (avy<jGestureResY.data[c1]) {
      avy=jGestureResY.data[c1];
      iny=c1;
    }
  }
  sprintf(jChar,"gesture Y %5.2f %5.2f %5.2f %5.2f",jGestureResY.data[0],jGestureResY.data[1],jGestureResY.data[2],jGestureResY.data[3]);
  jDeclared.prepend(new std::string(jChar));
  sprintf(jChar,"gesture X %5.2f %5.2f %5.2f %5.2f",jGestureResX.data[0],jGestureResX.data[1],jGestureResX.data[2],jGestureResX.data[3]);
  jDeclared.prepend(new std::string(jChar));
  if (inx==iny) {
    jGestureLast=inx+1;
    jSpeaks(jSPEAK_GESTURE);
    sprintf(jChar,"gesture %i detected",jGestureLast);
    jDeclared.prepend(new std::string(jChar));
  } else
    jGestureLast=jGESTURE_NONE;    
  return jGestureLast;
}

// used by pCode to obtain latest gesture
int jGestureQuery(void) {
  int tmp=jGestureLast;
  jGestureLast=jGESTURE_NONE;
  return tmp;
}

float jRad2Til(float rad) {
  return 0.27-1.155*rad;
}

// jLookAt is just roughly accurate
// pos is of course in local coordinates
bool jLookAt(float* pos,float height) {
  if (jTic[jTOC_NEXT_PTU]+1000<jTic[jTOC_LAST_SENSORS]) {
    matrix<float> cmp(2,1);
    cmp.data[0]=pos[0];
    cmp.data[1]=pos[1]-jPTU_OFFSET;
    float ang=atan2(-cmp.data[0],cmp.data[1]);          // angle for alignment (with respect to front, left is positive)
    float dst=cmp.norm(2);
    jSetPTU.data[0]=-ang*1.2f;                // MAGIC CONST
    jSetPTU.data[1]=jRad2Til(atan((height-jCAMERA_HEIGHT)/MAX(0.30f,dst)));
    return true;
  }
  return false;
}

bool jLookAt(float pan,float til) {
  if (jTic[jTOC_NEXT_PTU]+1000<jTic[jTOC_LAST_SENSORS]) {
    jSetPTU.data[0]=pan;
    jSetPTU.data[1]=til;
    return true;
  }
  return false;
}

bool jLookAround(float til) {
  if (jTic[jTOC_NEXT_LOOKAROUND]<jTic[jTOC_LAST_SENSORS]) {     // look around behavior
    jTic[jTOC_NEXT_LOOKAROUND]=jTic[jTOC_LAST_SENSORS]+2000+(int)(jRandu.token()*3000);
    jSetPTU.data[0]=(0.4+jRandu.token())*(1-2*(0<jOldPTU.data[0]));
    jSetPTU.data[1]=til;    
    return true;
  }
  return false;
}

