// code by jph
// included from ccode

// localizes marsokhod using sphere fit to raw laser points        (very accurate)
// alternatively minimum laser_point+laser_direction*teaser_radius (ok accurate)
void mLocalize(void) {
  jDEBUG("mLocalize")
  int c1,c2;
  float rad;
  matrix<float> tmp(2,1);
  matrix<float> tea=mPos*mOfs;
  matrix<float> als(3,30); als.numel=0;
  matrix<float> rhs(30,1); rhs.numel=0;
  matrix<float> dst(30,1); dst.numel=0;
  matrix<float> pnt=cPos*jLaserLocal;    
  for (c1=0;c1<pnt.size[1];c1++) {
    c2=c1*3;
    tmp.data[0]=pnt.data[c2+0]-tea.data[0];
    tmp.data[1]=pnt.data[c2+1]-tea.data[1];
    if (tmp.norm(2)<1.8f*mTEASER_RADIUS&&als.numel<90) {
      als.data[als.numel++]=2*pnt.data[c2+0];
      als.data[als.numel++]=2*pnt.data[c2+1];
      als.data[als.numel++]=1;
      rhs.data[rhs.numel++]=pnt.data[c2+0]*pnt.data[c2+0]+pnt.data[c2+1]*pnt.data[c2+1];
      dst.data[dst.numel++]=laser.distanceArray[c1].distance;
    }
  }
  als.size[1]=als.numel/3;
  rhs.size[0]=rhs.numel;
  dst.size[0]=dst.numel;
  if (3<rhs.numel) {
    matrix<float> mat=als*als.T();
    matrix<float> sol=mat.solve(als*rhs);      
    rad=sqrt(sol.data[2]+sol.data[0]*sol.data[0]+sol.data[1]*sol.data[1]);
    mMover->visual=jTic[jTOC_LAST_SENSORS]; // indicates that mhk is visible for j2b2
    Uint32 col=0xff8080ff;
    if (mTEASER_FITTOL<fabs(mTEASER_RADIUS-rad)&&jTic[jTOC_LAST_SENSORS]<mMover->stamp+jROLLOUT) { // use
      int ind;
      float cmp=100.0f;
      for (c1=0;c1<dst.numel;c1++)
        if (dst.data[c1]<cmp) {
          cmp=dst.data[c1];
          ind=c1;
        }
      matrix<float> ofs(3,1);
      ofs.data[0]=als.data[3*ind+0]*0.5f;
      ofs.data[1]=als.data[3*ind+1]*0.5f;
      ofs.data[2]=1;
      ofs=cInv*ofs;
      ofs.data[1]-=jLASER_OFFSET;
      ofs.data[2]=0;
      ofs=cPos*ofs;
      ofs.normalize(2);
      sol.data[0]=als.data[3*ind+0]*0.5f+ofs.data[0]*mTEASER_RADIUS;
      sol.data[1]=als.data[3*ind+1]*0.5f+ofs.data[1]*mTEASER_RADIUS;
      col=0xff20ffff;
      rad=mTEASER_RADIUS;
    }
    if (fabs(mTEASER_RADIUS-rad)<mTEASER_FITTOL) {
	    sol.data[0]+=pPos.data[6];
      sol.data[1]+=pPos.data[7];
      sol.data[2] =1;
      matrix<float> ref=mMover->pos.solve(sol);
      matrix<float> trf(3,3,"eye");
      float alp=ref.data[0]/mTEASER_OFFSET;
      float tau=ref.data[1]-mTEASER_OFFSET;
      if (fabs(alp)<0.3f&&fabs(tau)<0.3f) {
        trf.data[0]=trf.data[4]=cos(alp);
        trf.data[3]=sin(alp);
        trf.data[1]=-trf.data[3];
        trf.data[7]=tau;
        mMover->pos=mMover->pos*trf;
      }
    }
  }
  if (rhs.numel<6&&cDisplay==cDISPLAY_GEOGRAPHY) {
    sprintf(jChar,"%d",rhs.numel);
    jDrawString(cMap,jFonr,jChar,(int)(jMAP_X+tea.data[1]*jGND2MAP)-3,(int)(jMAP_Y+tea.data[0]*jGND2MAP)-4);
  }
  mPos=pInv*mMover->pos;
  mInv=cJustify_private(mPos);           // performs numerical rectification    
}

// determines best point/direction to follow marsokhod
// elegant solution because it uses tracking frames of mhk
void mStardusts(void) {
  jDEBUG("mStardusts")
  mDust.planned=false;
  if (!mMover->active)
    return;
  int c1,cnt=mStardust.size[1];
  int pix,piy;
  int cng=cnt;
  bool lan[cnt];
  bool dns[cnt];
  matrix<float> tok,pnt;
  for (c1=0;c1<cnt;c1++)
    lan[c1]=true;
  while (mMover->track.token(tok)) {
    pnt=pInv*tok*mStardust;
    for (c1=0;c1<cnt;c1++)
      if (lan[c1])
        lan[c1]&=cDensed(&pnt.data[c1*3]);
    for (c1=0;c1<cnt;c1++)
      if (lan[c1])
        if (dns[c1]=cDensed(cPos.aff_pnt(2),pnt.aff_pnt(c1)))
          cng=c1;
    if (cng<cnt) {
      if (jTic[jTOC_NEXT_STARDUST]<jTic[jTOC_LAST_SENSORS]||!(lan[jInt[jINT_STARDUST_LANE]]&&dns[jInt[jINT_STARDUST_LANE]])) {
        jTic[jTOC_NEXT_STARDUST]=jTic[jTOC_LAST_SENSORS]+8000;
        jInt[jINT_STARDUST_LANE]=cng;
      }
      mDust.planned=true;
      mDust.compass=pnt.aff_vec(jInt[jINT_STARDUST_LANE])-cPos.aff_vec(2);
      mDust.dist=mDust.compass.norm(2);
      mDust.compass.normalize(2);
      jDraw(jIN_MAP,&pnt.data[3*jInt[jINT_STARDUST_LANE]],1,0xff8040b0,&filledEllipseRGBA);
      break;
    }
  }
  mMover->track.next[2]=&mMover->track; // reset tokenizer
}

// function handles j2b2 behavior
void mFollowed(void) {  
  matrix<float> rel=(mInv*cPos).aff_vec(2);     // j2b2 with respect to marsokhod
  matrix<float> loc= mInv*mMover->plan.compass;
  bool flw=false;
  flw|=mSIZE0+jROBOT_RADIUS+0.3f<fabs(rel.data[0]);
  flw|=mSIZE1+jROBOT_RADIUS+0.3f<fabs(rel.data[1]);
  flw&=0<loc.data[1];                           // mhk is headed correctly
  jMover->active=rel.data[1]<0;                 // can still align but not backup
  if (jMover->active&&flw) {
    jTic[jTOC_STOP_FOLLOW]=jTic[jTOC_LAST_SENSORS]+3000;
    bool exc=false;
    float ang;
    matrix<float> com;
    if (exc=mDust.planned)                  // one of the lanes behind mhk
      com=cInv*mDust.compass;
    else
      if (exc=jMover->plan.planned)
        com=cInv*jMover->plan.compass;
    if (exc) {
      ang=atan2(-com.data[0],com.data[1]);                // angle for driving (with respect to front, left is positive)
      jSetSpeed.data[0]=0.75*jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang-fabs(jVel.data[0]));
      jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];          
      jLookAt((cInv*mPos.aff_pnt(2)).data,mSIZE2);
    } else 
      cTarget=mPos.aff_pnt(2);
  }
  if (!flw&&jTic[jTOC_STOP_FOLLOW]<jTic[jTOC_LAST_SENSORS]) {
    matrix<float> pos=(cInv*mPos).aff_vec(2);                   // marsokhod with respect to j2b2
    float hum=atan2(-pos.data[0],pos.data[1]);                  // angle for alignment (with respect to front, left is positive)
    jSetSpeed.data[1]=pGainAng*hum-pGainOme*jVel.data[0];
    if (20*jDEG2RAD<fabs(hum)||.1f<fabs(jVel.data[0]))
      jTic[jTOC_STOP_ALIGNING]=jTic[jTOC_LAST_SENSORS]+1500;
    if (jTic[jTOC_STOP_ALIGNING]<jTic[jTOC_LAST_SENSORS]&&pos.norm(2)<mSIZE1+jROBOT_RADIUS+0.2f) {       // if properly aligned backup
      matrix<float> prt=cLookInto(5,-cPos.data[3],-cPos.data[4]);
      if (0.25f+fabs(jVel.data[2])<prt.norm(2)) {               // if this is false, align!!!
        matrix<float> ret=cInv*prt;
        float ang=atan2( ret.data[0],-ret.data[1]);             // angle with respect to front, left is positive
        jSetSpeed.data[0]=-jSPEED0_SLW*MAX(0.0f,0.5f);
        jSetSpeed.data[1]= pGainInt*ang-pGainPos*jVel.data[0];
        jMover->active=1;
      }
    }
  }
}


// assumes mpath planned
// pos is in map coordinates
bool mCollides(float *pos,float *dir,float pl0,float pl1) {
  matrix<float> trf=mPos;
  trf.data[3]=dir[0];
  trf.data[4]=dir[1];
  trf.data[0]= trf.data[4];
  trf.data[1]=-trf.data[3];
  matrix<float> pnt(3,1);
  pnt.data[0]=pos[0]; pnt.data[1]=pos[1]; pnt.data[2]=1;
  pnt=trf.solve(pnt);
  return 0<pnt.data[1]&&pnt.data[1]<mSIZE1+pl1+jROBOT_RADIUS&&fabs(pnt.data[0])<mSIZE0+jROBOT_RADIUS+pl0;
}

// assumes mPlan is planned
// assumes mPos is NOT active => mActive
bool mAuroras(matrix<float>& yrp) {
  int c1,c2;
  int pix,piy,pin;
  float num;
  bool flg;
  matrix<float> tea=pInv*mMover->rested*mOfs;
  matrix<float> zen=pInv*mMover->rested*(mOfs*-1).aff_pnt(0);
  matrix<float> dir;
  if (jLocal2map(tea,pix,piy)) {                      // if teaser is in map
    matrix<float> srd=pInv*mMover->rested*mAurora;    // for drawing and cheking if reachable
    for (c1=0;c1<srd.size[1];c1++) {
      jLocal2map(srd.aff_pnt(c1),pix,piy);
      pixelRGBA(cMap,pix,piy,0x00,0x80,0x00,0xff);    // TODO REMOVE
    }
    array<float> srt(1,&mAurora.size[1],0.0f);
    for (c1=0;c1<mAurora.size[1];c1++)
      srt.data[c1]=2*mAurora.data[3*c1+1]+fabs(mAurora.data[3*c1+0]);
    matrix<float> tmp;
    array<int> ind=srt.sort();
    for (c1=0;c1<ind.numel;c1++) {
      tmp=srd.aff_pnt(ind.data[c1]);                  // in map coords
      if (jLocal2map(tmp,pix,piy)) {
        flg=true;
        pin=piy*jMAP_W+pix;
        for (c2=0;c2<16;c2++)        
          flg&=cRouteDist[pin+jRouteOfs[c2]]<cTERRAIN;      // first criteria: routable
        dir=tmp-tea;
        num=dir.norm(2)*jGND2MAP;
        for (c2=0;(c2<num)&&flg;c2++) {
          jLocal2map(tea+dir*(c2/num),pix,piy);
          flg&=cTerrain[piy*jMAP_W+pix]<=cTERRAIN_EOOB;  // second criteria: sight on teaser
        }
        dir=tmp-zen;
        num=dir.norm(2)*jGND2MAP;
        for (c2=0;(c2<num)&&flg;c2++) {
          jLocal2map(zen+dir*(c2/num),pix,piy);
          flg&=cTerrain[piy*jMAP_W+pix]<=cTERRAIN_EOOB;  // second criteria: sight on teaser
        }
        if (flg) {
          ellipseRGBA(cMap,pix,piy,4,4,0x20,0xf0,0x20,0xff);        
          yrp=tmp;
          return true;        
        }
      }
    }
  }
  return false;
}

bool mAuroras(void) {
  jSetSpeed*=0;
  sprintf(jChar,"aurora timeout %6.3f small %6.3f",(jTic[jTOC_STOP_AURORA]-jTic[jTOC_LAST_SENSORS])*0.001f,(jTic[jTOC_STOP_ALIGNING]-jTic[jTOC_LAST_SENSORS])*0.001f);
  jBulletin.append(new std::string(jChar));
  if (mJokerReg==1) {
    if (jMover->plan.planned&&mAuroras(cTarget)) {
      if (0.3f<jMover->plan.dist) {
        jTic[jTOC_STOP_ALIGNING]=jTic[jTOC_LAST_SENSORS]+2000;  // upper bound
        matrix<float> com=cInv*jMover->plan.compass;
        float ang=atan2(-com.data[0],com.data[1]);              // angle for driving (with respect to front, left is positive)
        jSetSpeed.data[0]=jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang-fabs(jVel.data[0])); // this works really nice!
        jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];          
      }
    } else 
      cTarget=cPos.aff_pnt(2);
    mJokerReg+=(jTic[jTOC_STOP_ALIGNING]<jTic[jTOC_LAST_SENSORS]||jTic[jTOC_STOP_AURORA]<jTic[jTOC_LAST_SENSORS]);
  }
  if (mJokerReg==0) { // criterion to trigger
    jTic[jTOC_STOP_AURORA]=jTic[jTOC_LAST_SENSORS]+35000;       // upper bound for driving along path bla
    jTic[jTOC_STOP_ALIGNING]=jTic[jTOC_LAST_SENSORS]+2000;      // upper bound
    jMover->active=1;                                           // j2b2 goes around
    mMover->active=0;                                           // park mhk
    jTic[jTOC_NEXT_NAVIGATION]=0;
    cTarget=cPos.aff_pnt(2);                                    // route from j2b2
    mJokerReg++;
  }
  if (mJokerReg==2) {                                           // alignment
    matrix<float> pos=(cInv*mPos).aff_vec(2);                   // marsokhod with resp to j2b2
    float hum=atan2(-pos.data[0],pos.data[1]);                  // angle for alignment (with respect to front, left is positive)
    jSetSpeed.data[1]=pGainAng*hum-pGainOme*jVel.data[0];
    if (20*jDEG2RAD<fabs(hum)||.1f<fabs(jVel.data[0]))
      jTic[jTOC_STOP_ALIGNING]=jTic[jTOC_LAST_SENSORS]+1500;
    if (jTic[jTOC_STOP_ALIGNING]<jTic[jTOC_LAST_SENSORS]) {
      mMover->active=1;
      mJokerReg=0;
    }
  }
  jSetSpeed*=!!mJokerReg;  
  return mJokerReg;
}

void mMobiles(void) {
  jDEBUG("mMobiles")
  int c1,c2;
  int pix,piy,pin;
  matrix<float> smp;
  for (c1=0;c1<mMOBILE;c1++) {
    mMobility.data[c1]=1;
    mJbotHaus.data[c1]=1;
    smp=mPos*mMobile[c1];
    for (c2=0;c2<smp.size[1]&&mMobility.data[c1];c2++)
      if (mMobility.data[c1]&=jLocal2map(smp.aff_pnt(c2),pix,piy)) 
        mMobility.data[c1]&=cTerrain[piy*jMAP_W+pix]<cTERRAIN_EOOB;
    smp=cInv*smp;
    for (c2=0;c2<smp.size[1];c2++) 
      mJbotHaus.data[c1]&=jROBOT_RADIUS+jAXIS_OFFSET+0.03f<smp.aff_vec(c2).norm(2);
    mMobility.data[c1]&=mJbotHaus.data[c1];
  }
  mMobility.data[mMOBILE_SPIN]&=mMobility.data[mMOBILE_LEFT]&&mMobility.data[mMOBILE_RIGHT];
}

// called from planner in pcode.cpp
// returns true if mhk makes progress
bool mTargets(void) {
  int c1;
  mSetSpeed*=0;
  if (mMover->plan.planned&&mMover->active&&mMover->visual==jTic[jTOC_LAST_SENSORS]) {
    if (!mCollides(&cPos.data[6],mMover->plan.compass.data,0.10f,0.25f)) {
      matrix<float> loc=mPos.solve(mMover->plan.compass);
      loc.data[2]=0; // should be zero anyways!!!
      float ang=atan2(-loc.data[0],loc.data[1]);    // angle with respect to front, left is positive
      float dst=loc.norm(2);
      if (0.12f<mMover->plan.dist) {
        mSetSpeed.data[1]=MIN(MAX(-mSPEED1_STD,-4*ang),mSPEED1_STD);
        mSetSpeed.data[1]*=mMobility.data[mMOBILE_RIGHT]||mSetSpeed.data[1]>0;
        mSetSpeed.data[1]*=mMobility.data[mMOBILE_LEFT ]||mSetSpeed.data[1]<0;
        mSetSpeed.data[0]=MIN(MAX(0.0f,10*loc.data[1]),mSPEED0_STD);
        mSetSpeed.data[0]=MAX(0.0f,mSetSpeed.data[0]-fabs(mSetSpeed.data[1]));
        mSetSpeed.data[0]*=mMobility.data[mMOBILE_FRONT];      
      } else
        sprintf(jChar,"mhk close to target");
    } else {
      sprintf(jChar,"mhk cannot target because location failure");
      jBulletin.append(new std::string(jChar));
    }
  }
  return 0<mSetSpeed.norm(2);
}

void mWiggleDefault(int deg) {
  float ang=deg*jDEG2RAD;
  mWiggle=matrix<float>(3,3);  
  ang*=1-2*(0.5<jRandu.token());
  mWiggle.data[0]=sin(ang);
  mWiggle.data[1]=cos(ang);
  mWiggle.data[3]=-mWiggle.data[0];    
  mWiggle.data[4]= mWiggle.data[1];
  mWiggle.data[6]=0;
  mWiggle.data[7]=1;
  mWiggle=mMover->pos*mWiggle;    
}

// j2b2 waits while Marsokhod spins left and right and left
// assumes matrix mWiggle is built
bool mWiggles(void) {
  if (mJokerReg==0) { // criterion to trigger    
    jMover->active=0;
    mMover->active=1;
    mJokerReg=1;
  }
  if (mJokerReg==1) { // move side 1
    matrix<float> loc=mInv*mWiggle.aff_vec(mWiggle.size[2]);
    float ang=atan2(-loc.data[0],loc.data[1]);    // angle with respect to front, left is positive
    jDraw(jIN_MAP,&mPos.data[6],mWiggle.aff_vec(mWiggle.size[2]).data,0xff000000,1);
    mSetSpeed.data[1]=MIN(MAX(-mSPEED1_STD,-5*ang),mSPEED1_STD);
    mSetSpeed.data[1]*=mMobility.data[mMOBILE_RIGHT]||mSetSpeed.data[1]>0;
    mSetSpeed.data[1]*=mMobility.data[mMOBILE_LEFT ]||mSetSpeed.data[1]<0;
    sprintf(jChar,"wiggle step %d angle %6.3f vel %6.3f",mWiggle.size[2],ang,mSetSpeed.data[1]);
    jBulletin.prepend(new std::string(jChar));
    if (0.35f<fabs(mSetSpeed.data[1])) // XX deg precision
      jTic[jTOC_NEXT_JOKER_REGISTER]=jTic[jTOC_LAST_SENSORS]+1000;
    if (jTic[jTOC_NEXT_JOKER_REGISTER]<jTic[jTOC_LAST_SENSORS])
      mWiggle.size[2]++;                // TODO use delayed instead!!!
    mJokerReg*=mWiggle.size[2]<mWiggle.size[1];
  }
  return mJokerReg;
}

bool mBackups(void) {
  jDEBUG("mBackups")
  float pce=(cInv*mPos*mOfs).aff_vec(0).norm(2);
  if (mJokerReg==0&&pce<2) { // criterion to trigger    
    jTic[jTOC_STOP_BACKUP]=jTic[jTOC_LAST_SENSORS]+1500;
    jTic[jTOC_NEXT_NAVIGATION]=0;
    jMover->active=1;
    mJokerReg=1;
  }
  if (mJokerReg==1) {
    matrix<float> tmp=mPos*mOfs;
    matrix<float> prt=cLookInto(15,cPos.data[6]-tmp.data[0],cPos.data[7]-tmp.data[1]);
    if (0.25f+fabs(jVel.data[2])<prt.norm(2)) {               // if this is false, align!!!
      jTic[jTOC_STOP_BACKUP]=jTic[jTOC_LAST_SENSORS]+1500;
      matrix<float> ret=cInv*prt;
      float ang=atan2( ret.data[0],-ret.data[1]);             // angle with respect to front, left is positive
      jSetSpeed.data[0]=-jSPEED0_SLW*MAX(0.0f,0.5f-2*ang*ang);
      jSetSpeed.data[1]= pGainInt*ang-pGainPos*jVel.data[0];
    } else
      jSetSpeed*=0;
    if (jTic[jTOC_STOP_BACKUP]<jTic[jTOC_LAST_SENSORS]||2<pce) {
      jMover->active=0;
      jSetSpeed*=0;
      mJokerReg=0;
    }
  }
  sprintf(jChar,"backup %d %6.3f",jTic[jTOC_LAST_SENSORS]-jTic[jTOC_STOP_BACKUP],pce);
  jBulletin.prepend(new std::string(jChar));
  return mJokerReg;
}

void mJokers(void) {
  int c1;
  mJokerReg=0;
  for (c1=0;c1<mJOKER;c1++)
    mJokerPos[c1]=matrix<float>(3,3,"eye");
}

bool mJokers(bool launch,int type) {
  jDEBUG("mJokers")
  int c1;
  bool exc=false;
  if (launch&&type&&!mJokerReg) {
    mJoker=mJOKER;
    for (c1=0;c1<mJOKER;c1++)  // fill distance tables
      type&=~((mMover->pos.solve(mJokerPos[c1]).aff_vec(2).norm(2)<mJOKER_RADIUS)<<c1); // this is clearly insane
    sprintf(jChar,"type %d execute %d joker %d reg %d",type,exc,mJoker,mJokerReg);
    jDeclared.prepend(new std::string(jChar));        
    matrix<float> m2j=mInv*cPos.aff_pnt(2); // j2b2 with respect to marsokhod
    if (m2j.data[1]<0&&(mSIZE0+jROBOT_RADIUS+0.2f<fabs(m2j.data[0])||mSIZE1+jROBOT_RADIUS+0.2f<fabs(m2j.data[1]))) {    // j2b2 behind marsokhod
      jDeclared.prepend(new std::string("selecting joker: wiggle"));    
      if (type&mJOKER_WIGGLE_)      { mJoker=mJOKER_WIGGLE;     mWiggleDefault(35); }
      if (type&mJOKER_TINYWIGGLE_)  { mJoker=mJOKER_TINYWIGGLE; mWiggleDefault(20); }
    } else {                // j2b2 in front of marsokhod
      jDeclared.prepend(new std::string("selecting joker: backup, aurora"));
      if (type&mJOKER_AURORA_)      mJoker=mJOKER_AURORA;
      if (type&mJOKER_BACKUP_)      mJoker=mJOKER_BACKUP;
    }
    if (mJoker<mJOKER)
      mJokerPos[mJoker]=mMover->pos;
  }
  if (mJoker==mJOKER_BACKUP)
    exc=mBackups();
  if (mJoker==mJOKER_AURORA)
    exc=mAuroras();
  if (mJoker==mJOKER_WIGGLE)  
    exc=mWiggles();
  if (mJoker==mJOKER_TINYWIGGLE)
    exc=mWiggles();
  if (!exc) {
    mJoker=mJOKER;
    mJokerReg=0;  
  }
  return exc;
}
