// code by jph

// accumulates raw laser points to recording geometry
// resolution 2 cm
void pRecord(void) {
  if (!jTog[jTOG_GLOBAL_RECORD]) return;
  int c1;
  matrix<float> frm=pPos*cPos;
  // BEGIN: store robots state and timestamp
  int ptr;
  ptr=pTrace.numel;
  pTrace.data[ptr+0]=frm.data[0];
  pTrace.data[ptr+1]=frm.data[1];
  pTrace.data[ptr+2]=frm.data[3];
  pTrace.data[ptr+3]=frm.data[4];
  pTrace.data[ptr+4]=frm.data[6];
  pTrace.data[ptr+5]=frm.data[7];
  pTrace.data[ptr+6]=jTic[jTOC_LAST_SENSORS]*0.001;
  pTrace.size[1]++;
  pTrace.size[1]%=pTRACE;
  pTrace.numel=pTrace.size[0]*pTrace.size[1];
  // END: store robots state and timestamp
  // BEGIN: hash solid geometry
  matrix<float> tmp(2,1);
  matrix<float> pnt=frm*jLaserLocal;
  matrix<int> ind(2,1);
  for (c1=0;c1<pnt.numel;c1+=3) {
    tmp.data[0]=jLaserLocal.data[c1+0];
    tmp.data[1]=jLaserLocal.data[c1+1];
    if (tmp.norm(2)<5.0f) {
      tmp.data[0]=pnt.data[c1+0];
      tmp.data[1]=pnt.data[c1+1];
      ind.data[0]=MAX(0,(int)(cSOLID_OFFSET+jGND2MAP*tmp.data[0]));
      ind.data[1]=MAX(0,(int)(cSOLID_OFFSET+jGND2MAP*tmp.data[1]));
      cSolid.add(ind);
    }
  }
  sprintf(jChar,"[REC] Trace %i, Solids %i (Press s to save)",pTrace.size[1],cSolid.numel);
  jBulletin.append(new std::string(jChar));  
  // END: hash solid geometry
}

// dumps geometry collection into binary file with coordinates
// and bitmaps of points
void pRecordStore(void) {
  if (!jTog[jTOG_GLOBAL_RECORD]||cSolid.numel==0)
    return;
  sprintf(jChar,"sav/yourtrace%i.bin",pTrace.size[1]);
  pTrace.store(jChar);
  // BEGIN: compile geometry into one array
  int c1,cnt=0;    
  int dim[2]={2,cSolid.numel};
  int miw= 10000000,mih= 10000000;
  int maw= 0,mah= 0;
  array<int> arr(2,dim,0);
  matrix<int> tok;
  for (c1=0;c1<cSOLID;c1++)
	  while (cSolid.table.data[c1]->token(tok)) {
	    arr.data[cnt++]=tok.data[0];
	    arr.data[cnt++]=tok.data[1];
	    miw=MIN(miw,tok.data[1]); mih=MIN(mih,tok.data[0]);
	    maw=MAX(maw,tok.data[1]); mah=MAX(mah,tok.data[0]);
    }
  sprintf(jChar,"sav/yoursolid%i.set",cSolid.numel);
  arr.store(jChar);  
  // END: compile geometry into one array  
  // BEGIN: compile geometry into one bitmap  CRASHES AND/OR DOES NOT PRODUCE READIBLE FILE!!!
  if (!pAtlas) return;
  if (!pAtlas->format->palette) return;
  maw=maw-miw+1;
  mah=mah-mih+1;
  maw/=8;
  ++maw;
  maw*=8;
  SDL_Surface* fld=SDL_CreateRGBSurface(SDL_SWSURFACE,maw,mah,8,0x00,0x00,0x00,0x00);
  SDL_LockSurface(fld);
  fld->format->palette=pAtlas->format->palette;
  Uint8* ptr=(Uint8*)(fld->pixels);
  for (c1=0;c1<cSOLID;c1++)
	  while (cSolid.table.data[c1]->token(tok))
	    ptr[(tok.data[0]-mih)*maw+tok.data[1]-miw]=pATLAS_WALL;
  SDL_UnlockSurface(fld);
  sprintf(jChar,"sav/youratlas%i.bmp",cSolid.numel);
  SDL_SaveBMP(fld,jChar);
	fld->format->palette=0;
  SDL_FreeSurface(fld);
  jBulletin.append(new std::string("stored recorded solids"));
  // END: compile geometry into one bitmap
}


void pManage() {
  jDEBUG("pManage")
  matrix<float> sta;
  jMoviton* tok;
  while (pMoves.token(tok)) {
    if (tok->active) {
      tok->rested=tok->pos;
      cPlanning(tok);  // planning in local coordinates and updated every iteration      
    }
    sta=tok->track.next[0]->data;       // get first element from list
    sta.data[0]=sta.data[1]=sta.data[3]=sta.data[4]=sta.data[8]=0;
    sta.data[6]-=tok->pos.data[6];
    sta.data[7]-=tok->pos.data[7];
    if (pMOVES_TRACK<sta.norm(2))
      tok->track.prepend(tok->pos);
  }
}

// updates the moviton* list
// all cMoves are within the FREE terrain of the atlas, guaranteed!
void pPlaces(linked<jMoviton*>& plc,matrix<float> cmv,bool& jmd,float psp) {
  jDEBUG("pPlaces")
  int c1=pPLACE,c2,cnt=0;
  jmd=false;                          // pretend no event
  if (!cmv.size[1])
    return;
  array<jMoviton*>  hum(1,&c1,NULL);
  array<int>        asc(1,&cmv.size[1],-1);
  jMoviton* tok;
  for (c1=0;c1<cmv.numel;c1+=3) {
    cmv.data[c1+0]+=pPos.data[6];     // transform movements into global system
    cmv.data[c1+1]+=pPos.data[7];
  }
  while (plc.token(tok))             // clear linked list, and store into array
    if (tok->id<=jMOVITON_ID_HUMAN) {
      plc.remove();                     // does not delete the content
      tok->add*=0;                      // internal const
      hum.data[cnt++]=tok;
    }  
  // BEGIN: associate each move to the most appropriate guy, index abandoned moves
  int pix,piy;
  float bst;
  linked<int> add;                    // stores what?
  matrix<float> tmp(2,1);
  for (c1=0;c1<cmv.size[1];c1++) {    // compare every perceived motion
    bst=psp;                          //        require 50 cm to match perceived motion to existing human
    for (c2=0;c2<cnt;c2++) {          // => with every existing human
      tmp.data[0]=cmv.data[c1*3+0]-hum.data[c2]->pos.data[6];
      tmp.data[1]=cmv.data[c1*3+1]-hum.data[c2]->pos.data[7];
      if (tmp.norm(2)<bst) {
        bst=tmp.norm(2);
        asc.data[c1]=c2;              // carries the best human to cmove
      }
    }
    if (asc.data[c1]<0)               // move abandoned 
      add.append(c1);                 // this is bounded to up by radius of diff, cmove c1 is associated to human asc.data[c1]            
  }
  // END: associate each move to the most appropriate guy, index abandoned moves
  // BEGIN: distribute the moves to the best guys (average)
  for (c1=0;c1<cmv.size[1];c1++) {
    c2=asc.data[c1];
    if (0<=c2) {                      // if no human associated to cmove c2 is -1
      hum.data[c2]->add.data[0]+=cmv.data[c1*3+0]*cmv.data[c1*3+2];
      hum.data[c2]->add.data[1]+=cmv.data[c1*3+1]*cmv.data[c1*3+2];
      hum.data[c2]->add.data[2]+=cmv.data[c1*3+2];
    }
  }
  matrix<float> pqs(3,1);             // to store old position
  for (c1=0;c1<cnt;c1++) {            // for all humans
    tok=hum.data[c1];                 // get pointer to human
    if (jEPS<tok->add.data[2]) {
      jmd=true;
      tok->stamp=jTic[jTOC_LAST_SENSORS];
      pqs.data[0]=tok->pos.data[6];
      pqs.data[1]=tok->pos.data[7];
      tok->add*=1/tok->add.data[2];   // make real new position
      tok->pos.data[6]=tok->add.data[0];  // old position blended with new
      tok->pos.data[7]=tok->add.data[1];
      tmp.data[0]=tok->pos.data[6]-pqs.data[0];    // new position - old position
      tmp.data[1]=tok->pos.data[7]-pqs.data[1];    // new position - old position
      tmp.normalize(2);
      tok->pos.data[3]= tmp.data[0];
      tok->pos.data[4]= tmp.data[1];
      tok->pos.data[0]= tmp.data[1];
      tok->pos.data[1]=-tmp.data[0];
    }
  }
  // END: distribute the moves to the best guys
  // BEGIN: merge two close ups OR remove
  bool app;
  int nws=0;
  for (c1=0;c1<cnt;c1++)                                              // for all humans
    if (tok=hum.data[c1]) {
      app=true;                                                       // assume everything is fine
      for (c2=c1+1;c2<cnt;c2++)
        if (hum.data[c2]) {
          tmp.data[0]=tok->pos.data[6]-hum.data[c2]->pos.data[6];
          tmp.data[1]=tok->pos.data[7]-hum.data[c2]->pos.data[7];
          app&=psp<tmp.norm(2);                                       // all others far enough away TODO AVG!?!?
        }
      if (app||tok->id) {                       // move survives and is appended once more
        plc.append(tok);                                              // join to list once more
        nws++;
      } else {
        delete hum.data[c1];
        hum.data[c1]=0;
      }
    }
  // END: merge two close ups OR remove 
  matrix<float> pos(3,1);                               // JUST A REFERENCE POINT
  // BEGIN: add a single new guy  
  bool fin;
  if (add.token(c1)&&nws<pPLACE) {                      // c1 is the cmove CAREFUL with pMover!!!
    pos.data[0]=cmv.data[c1*3+0];
    pos.data[1]=cmv.data[c1*3+1];
    fin=true;
    while (add.token(c1)) {
      tmp.data[0]=cmv.data[c1*3+0]-pos.data[0]; // compare cmove c1 with all the others cmoves
      tmp.data[1]=cmv.data[c1*3+1]-pos.data[1]; // guarantee that cmvs are close together
      fin&=tmp.norm(2)<psp;
    }
    if (fin) {
      jmd=true;
      tok=new jMoviton;
      tok->id    =jMOVITON_ID_UNKNOWN;
      tok->stamp =jTic[jTOC_LAST_SENSORS];
      tok->visual=jTic[jTOC_LAST_SENSORS];
      tok->pos=matrix<float>(3,3,"eye");
      tok->pos.data[8]=0;
      while (add.token(c1)) {                   
        tok->pos.data[6]+=cmv.data[c1*3+0]; // since it is guaranteed that cmvs are close together
        tok->pos.data[7]+=cmv.data[c1*3+1]; // we compute average now
        tok->pos.data[8]+=1.0f;
      }
      tok->pos.data[6]/=tok->pos.data[8];
      tok->pos.data[7]/=tok->pos.data[8];
      tok->pos.data[8] =1.0f;
      tok->track.prepend(tok->pos);          // add place of birth
      tok->ini=matrix<float>(3,1);
      tok->ini.data[0]=tok->pos.data[6];
      tok->ini.data[1]=tok->pos.data[7];
      tok->ini.data[2]=1;
      tok->add=matrix<float>(3,1);
      tok->contour=&pHuman;      
      tok->active=1;
      tok->rested=tok->pos;
      plc.append(tok);
    }
  }
  // END: add a single new guy
}

matrix<float> pMovesAura(matrix<float> glb) {             // glb is the 3x3 global coordinates matrix of a human, e.g.
  glb.data[6]-=pPos.data[6];
  glb.data[7]-=pPos.data[7];
  matrix<float> trf=cInv*glb;
  float ang=atan2(-trf.data[6],trf.data[7]);              // angle with respect to front, left is positive
  trf.data[0]= cos(ang); trf.data[3]=-sin(ang);
  trf.data[1]= sin(ang); trf.data[4]= cos(ang);
  return cPos*trf*cMovesAura;
}

void pMovesUpon(void) {
  jDEBUG("pMovesUpon")
  int c1,c2;
  bool flg=false;
  SDL_LockSurface(cTerrainHoly);
  Uint32* myHoly=(Uint32*)(cTerrainHoly->pixels);         // to manipulate pixel array
  // BEGIN: free the terrain in the back of the moving human (but only if terrain has ever been FREE)
  matrix<float> pos(3,1);
  SDL_FillRect(cTerrainStub,NULL,0);                      // clear stub
  jMoviton* tok;
  matrix<float> sta;
  while (pMoves.token(tok))
    if (tok->id==jMOVITON_ID_HUMAN&&tok->stamp==jTic[jTOC_LAST_SENSORS]) { // this means human has just moved
      flg=true;
      jDraw(jIN_TERRAINSTUB,pMovesAura(tok->pos)                ,0xffffffff,&filledPolygonRGBA); //  XX
    }
  if (flg) {
    SDL_LockSurface(cTerrainStub);
    Uint32* myStub=(Uint32*)(cTerrainStub->pixels);       // to manipulate pixel array
    for (c1=0;c1<cTERRAIN;c1++)
      if (myHoly[c1]&&myStub[c1])
        cTerrain[c1]=cTERRAIN_FREE;
    SDL_UnlockSurface(cTerrainStub);
  }
  // END: free the terrain in the back of the moving human
  // BEGIN: classify visibility of humans
  int pix,piy;
  matrix<float> lag=jMover->pos*jLaserLocal;
  matrix<float> tmp;
  while (pMoves.token(tok))                  // for all humans
    if (tok->id<=jMOVITON_ID_HUMAN) {
      tmp=tok->pos.solve(lag);
      for (c1=0;c1<tmp.size[1];c1++)
        if (tmp.aff_vec(c1).norm(2)<pMOVES_VISUAL) {
          tok->visual=jTic[jTOC_LAST_SENSORS];
          break;
        }
    }
  // END: classify visibility of humans
  SDL_UnlockSurface(cTerrainHoly);
  if (jTog[jTOG_MOTION_DETECTED])                         // from last iteration
    cROUTE_FREQUENT;
}

// humans manager
// human master manager
void pMovers(void) {
  jDEBUG("pMovers")
  jMoviton* tok;
  jMoviton* alt=pMover;                              // alternative to present human master
  while (pMoves.token(tok))
    switch (tok->id) {
      case jMOVITON_ID_UNKNOWN:
        if (pMOVES_WANDER<tok->pos.solve(tok->ini).aff_vec(0).norm(2))
          tok->id=jMOVITON_ID_HUMAN;
        if (tok->stamp+1000<jTic[jTOC_LAST_SENSORS]||tok->visual!=jTic[jTOC_LAST_SENSORS]) {
          pMoves.remove();
          delete tok;
        }
        break;
      case jMOVITON_ID_HUMAN:
        if (tok->stamp<jTic[jTOC_LAST_SENSORS]-pMOVES_GRAVEYARD||(jMover->pos.solve(tok->pos.aff_pnt(2)).data[1]<0.3f&&tok->stamp<jTic[jTOC_LAST_SENSORS]-1000)) {
          if (tok==pMover) {
            if (alt==pMover)
              alt=jMover;
            pMover=jMover;
          }
          pMoves.remove();
          delete tok;
        } else 
          if (tok->stamp==jTic[jTOC_LAST_SENSORS]) {
            if (pMover->id!=jMOVITON_ID_HUMAN) {      // pMover is free and tok has just moved
              jSpeaks(jSPEAK_WELCOME);
              alt=tok;                                // declare human as pMover
            } else
              if (pMover->stamp+3000<jTic[jTOC_LAST_SENSORS])
                alt=tok;                              // after non motion switch to moving human
          }
        break;      
    }
  if (jTog[jTOG_LOOKING_FOR_HUMAN]) {
    sprintf(jChar,"looking for human master");
    jBulletin.append(new std::string(jChar));
    pMover=alt;
  }
}



void pAtlasConv(void) {
  int c1,c2;
  int w=pAtlas->w;
  int ind[9]={-w-1,-w+0,-w+1,       -1,   0,  +1,       w-1, w+0, w+1};
  int num=w*pAtlas->h-w-1;
  int sum;
  SDL_LockSurface(pAtlas);
  Uint8* ptr=(Uint8*)(pAtlas->pixels);
  for (c1=w+1;c1<num;++c1) {
    if (ptr[c1]==pATLAS_WALL) {
      sum=0;
      for (c2=0;c2<9;c2++)
        sum+=(ptr[c1+ind[c2]]&pATLAS_WALL)==pATLAS_WALL;
      if (6<sum)
        ptr[c1]|=pATLAS_INJC;
    }
  }
  for (c1=w+1;c1<num;++c1) {
    if (ptr[c1]&pATLAS_WALL)
      if (ptr[c1]&pATLAS_INJC)
        ptr[c1]=pATLAS_WALL;
      else
        ptr[c1]=pATLAS_VOID;
  }
  SDL_UnlockSurface(pAtlas);
}

void pAltar(void) {
  int c1;
  if (jTog[jTOG_ATLAS_CALIBRATED]&&pAtlas) {              // atlas calib OR blit    
    SDL_FillRect(cTerrainStub,NULL,0);
    SDL_FillRect(cTerrainHoly,NULL,0);
    SDL_Rect atl;
  	atl.x=(int)(jMAP_X-pPos.data[7]*jGND2MAP);
  	atl.y=(int)(jMAP_Y-pPos.data[6]*jGND2MAP);
  	SDL_BlitSurface(pAtlas,NULL,cTerrainStub,&atl);
  	SDL_LockSurface(cTerrainStub);
  	SDL_LockSurface(cTerrainHoly);
  	Uint32* myStub=(Uint32*)(cTerrainStub->pixels);       // to manipulate pixel array
  	Uint32* myHoly=(Uint32*)(cTerrainHoly->pixels);       // to manipulate pixel array
  	cAtlasNum=0;
  	cDrainNum=0;
  	for (c1=0;c1<cTERRAIN;c1++)
  	  switch (myStub[c1]) {
    	  case 0xff008000:
  	      cAtlasInd[cAtlasNum++]=c1;
  	      break;
    	  case 0xff808080:
  	      if (cTerrain[c1]==cTERRAIN_NOGO)
  	        cTerrain[c1]=cTERRAIN_FREE;
  	      myHoly[c1]=cTERRAIN_FREE|0xffd0d0d0;            // d = 13, so what
  	      break;
  	    case 0xff800000:
  	      cDrainInd[cDrainNum++]=c1;                      // no trespassing
  	      break;
	    }
  	SDL_UnlockSurface(cTerrainStub);
  	SDL_UnlockSurface(cTerrainHoly);
    for (c1=0;c1<cAtlasNum;c1++)                          // make localization snap to terrain from atlas data
      cTerrain[cAtlasInd[c1]]|=cTERRAIN_WALL;
  }
  jTog[jTOG_ATLAS_CALIBRATED]=false;
}

int pRouting(float* pnt) {
  int c1;
  int pix,piy;
  pix=(int)(jGND2SNK*pnt[1]);
  piy=(int)(jGND2SNK*pnt[0]);
  matrix<float> dst(2,1);
  int bst=0;
  float cmp,low=1000;
  for (c1=pSNEAK_SZE;c1<pSneakNum;c1+=pSNEAK_SZE) {
    dst.data[0]=pix-pSneakInt[c1+pSNEAK_PIX];
    dst.data[1]=piy-pSneakInt[c1+pSNEAK_PIY];
    cmp=dst.norm(2);
    if (cmp<low) {
      low=cmp;
      bst=c1;
    }
  }
  return bst;
}

// needs to be input!
void pTracking(void) { // use pMover!
  jDEBUG("pTracking")
  if (!jTog[jTOG_HEAVY_DUTY]&&
      (jTic[jTOC_NEXT_TRACKING]<jTic[jTOC_LAST_SENSORS]||0.3f<(pTarget-pTargetOld).norm(2)||pMoverOld!=pMover)) {
    int c1;
    int pix,piy,pin;
    int ori,tag,tok,nxt;
    ori=pRouting(&pMover->pos.data[6]);                   // current robot approximate position
    jDEBUG("pMover not null, yippee!")
    tag=pRouting(pTarget.data);                           // target approximate position
    pSneakInt[pSNEAK_PRE]=1;                              // initialization, dirty trick: noNeighbor is alread visited
    for (c1=pSNEAK_SZE;c1<pSneakNum;c1+=pSNEAK_SZE)
      pSneakInt[c1+pSNEAK_PRE]=0;                         // initialization
    pSneakInt[tag+pSNEAK_PRE]=tag;                        // self reference
    int stack[30];
    int head=0,tail=0;
    stack[head++]=tag;
    while (tail<head) {
      nxt=stack[tail++];
      do {
        tok=nxt;
        nxt=0;
        for (c1=0;c1<3;c1++)                              // at most three neighbors
          if (!pSneakInt[pSneakInt[tok+c1]+pSNEAK_PRE]) { // neighbor unvisited?
            pSneakInt[pSneakInt[tok+c1]+pSNEAK_PRE]=tok;
            if (!nxt)                                     // no neighbor in pipeline
              nxt=pSneakInt[tok+c1];                      // mark neighbor as next
            else
              stack[head++]=pSneakInt[tok+c1];            // put neighbor to list
          }
      } while (nxt);
    }
    pSneakInt[tag+pSNEAK_PRE]=tag;                        // just to be on the safe side?
    matrix<float> pos;
    while (pTrack.token(pos))
      pTrack.remove();    
    nxt=ori;
    pos=matrix<float>(3,1);
    pos.data[2]=1;
    jInt[jINT_TRACKING_LENGTH]=0;
    jInt[jINT_TRACKING_SEGMENT]=0;
    int icr=1;
    if (pSneakInt[nxt+pSNEAK_PRE]) {                      // this is false when ori and target not in same connected component
      while (nxt!=pSneakInt[nxt+pSNEAK_PRE]) {
        pos.data[0]=pSNK2GLB0(nxt);
        pos.data[1]=pSNK2GLB1(nxt);
        pTrack.append(pos);
        icr&=!pSneakInt[nxt+2];
        jInt[jINT_TRACKING_LENGTH]++;        
        jInt[jINT_TRACKING_SEGMENT]+=icr;
        nxt=pSneakInt[nxt+pSNEAK_PRE];                
      }
      pTrack.append(pTarget);                               // final point
      jInt[jINT_TRACKING_LENGTH]++;
    }
    jTic[jTOC_NEXT_TRACKING]=jTic[jTOC_LAST_SENSORS]+2000;
    jTog[jTOG_HEAVY_DUTY]=true;
    pTargetOld=pTarget;
    pMoverOld =pMover;
   	jDrawString(screen,jFont,"T",cICON_TRACKING_X,cICON_Y);
   	jTog[jTOG_FRESH_TRACKING]=true;
  } else {
    jDrawString(screen,jFond,"T",cICON_TRACKING_X,cICON_Y);
    jTog[jTOG_FRESH_TRACKING]=false;
  }
}


void pRoundtrip(void) {
  int c1,c2,ind=0;
  float dst,cmp=10000;
  matrix<float> pos(3,1);
  for (c1=0;c1<=pLANDER_ROBOTSHEAVEN;c1++) {
    pos.data[0]=pPos.data[6]+cPos.data[6]-pLander.data[2*c1+1]*jSNK2GND;
    pos.data[1]=pPos.data[7]+cPos.data[7]-pLander.data[2*c1+0]*jSNK2GND;
    dst=pos.norm(2);
    if (dst<cmp) {
      cmp=dst;
      ind=c1;
    }
  }
  for (c1=0;c1<=pLANDER_ROBOTSHEAVEN;c1++) {
    c2=c1+ind;
    c2%=pLANDER_ROBOTSHEAVEN+1;
    pos.data[0]=pLander.data[2*c2+1]*jSNK2GND;
    pos.data[1]=pLander.data[2*c2+0]*jSNK2GND;
    pos.data[2]=1;
    pTargetList.append(pos);
    sprintf(jChar,"%i landmark %s",c1,pLanderLabel[c2].c_str());
    jDeclared.prepend(new std::string(jChar));
  }
}

