// code by jph

#define                 pSTATE_IDLE                       0
#define                 pSTATE_PURSUIT                    1         // pursue in direction of drag
#define                 pSTATE_SELECT_TARGET              2         // bumper press selects one of 4 points in lab as destination
#define                 pSTATE_TRAIN_MARKER               3         // calibrate palette
#define                 pSTATE_GUIDANCE                   4         // robot drives to list of targets in lab
#define                 pSTATE_AFFECTION                  5         // robot follows human
#define                 pSTATE_RECHARGE                   6         // dignified waiting
#define                 pSTATE_REMOTE_CONTROL             7         // bumper control marsokhod
#define                 pSTATE_MBOT_CIRCLE                10        // drive to spot where a 180 turn is possible
#define                 pSTATE_MBOT_INTERPLAY             11        // j2b2 follows marsokhod
#define                 pSTATE_MBOT_DEBUG                 12        // 

#define                 pSTATE_DRIVE_ALONG_PATH           18
#define                 pSTATE_DEBUG                      19

#define                 pSTATE_TRY(ms)                    jTic[jTOC_NEXT_STATE]=jTic[jTOC_LAST_SENSORS]+ms;
#define                 pSTATE_GIVEUP                     (jTic[jTOC_NEXT_STATE]<jTic[jTOC_LAST_SENSORS])||(jTic[jTOC_NEXT_RECHARGE]<jTic[jTOC_LAST_SENSORS])
#define                 pSTATE_RESET                      pStates[pIndex=0]=pSTATE_IDLE; pRegister=0; jTic[jTOC_NEXT_STATE]=0
                        
#define                 pPLACE                            15        // maximum number of located items

#define                 pSTONE_SPACING                    0.35f
#define                 pMOVES_SPACING                    0.50f     // at least 2x human radius, but why not a little more, to id human even when walking behind small geometry
#define                 pMOVES_VISUAL                     0.18f     // if human "visible" laser points should be within this range from pos
#define                 pMOVES_WANDER                     0.35f     // the human should have traveled
#define                 pMOVES_TRACK                      0.10f
#define                 pMOVES_GRAVEYARD                  20000     // terminate human after X sec of non-motion
#define                 pMOVER_VISUAL                     (jTic[jTOC_LAST_SENSORS]<pMover->visual+800)

#define                 pPTU_GRACEFUL                     (jTic[jTOC_NEXT_PTU]+1000<jTic[jTOC_LAST_SENSORS])
#define                 pPTU_GRANDMOTHER                  (jTic[jTOC_NEXT_PTU]+2000<jTic[jTOC_LAST_SENSORS])
#define                 pSPEED_GOGETEMTIGER               if (1.8f<jMover->plan.safe&&fabs(jVel.data[0])<0.1&&fabs(ang)<0.1) { jSetSpeed.data[0]=0.3; jBulletin.append(new std::string("safe to tiger")); jTic[jTOC_NEXT_SCREENSHOT]=MAX(jTic[jTOC_NEXT_SCREENSHOT],jTic[jTOC_LAST_SENSORS]+10000); }
// pSPEED_GOGETEMTIGER needs a good framerate otherwise localization screws up

#define                 pTRACK_RADIUS                     0.125f    // distance between drop off points
                        
#define                 pDIST_FOLLOW                      1.6f+0.15f*(pMover->id==jMOVITON_ID_MARSOKHOD)
#define                 pDIST_BACKUP                      0.9f
#define                 pDIST_REAL3D                      0.6f-1.15f*(pMover->id==jMOVITON_ID_MARSOKHOD)
#define                 pDIST_WINDOW                      35*jDEG2RAD
#define                 pDIST_PURSUIT                     0.15f      // min distance carried so that J2B2 will pursue direction

// BEGIN: globals
int                     pStates[20];                                // sequence of elementary states (max 20) !!! NO init for this!?!?!
int                     pIndex=0;
int                     pRegister=0;
matrix<float>           pPursue(3,1);                               // direction of pursue mode
linked<matrix<float> >  pTargetList;
matrix<float>           pTarget(3,1);                               // world coordinates
matrix<float>           pTargetOld(3,1);                            // point of last routing source
linked<matrix<float> >  pTrack;
#define                 pTRACE                            3600      // space allocated to record consecutive frames
matrix<float>           pTrace(7,pTRACE);

#define                 pLANDER                           5
#define                 pLANDER_CENTERALPARK              0
#define                 pLANDER_BIRTHPLACE                1         // in pixel coordinates
#define                 pLANDER_COFFEEWELL                2 
#define                 pLANDER_ROBOTSHEAVEN              3         // also: last location in roundtrip
#define                 pLANDER_TOXICWASTE                4
matrix<int>             pLander(2,pLANDER);
std::string             pLanderLabel[pLANDER];

#define                 pSNEAK                            16384
#define                 pSNEAK_PRE                        3
#define                 pSNEAK_PIX                        4
#define                 pSNEAK_PIY                        5
#define                 pSNEAK_PIN                        6
#define                 pSNEAK_SZE                        8
#define                 pSNK2GLB0(c1)                     (pSneakInt[c1+pSNEAK_PIY]*jSNK2GND+0.04f)
#define                 pSNK2GLB1(c1)                     (pSneakInt[c1+pSNEAK_PIX]*jSNK2GND+0.04f)


SDL_Surface*            pAtlas=0;
SDL_Surface*            pSneak=0;
SDL_Surface*            cFlag=0;
SDL_Surface*            pFlag=0;
int                     pSneakInt[pSNEAK];
int                     pSneakNum=0;
matrix<float>           pHuman(3,10);
jMoviton*               pMoverOld=0;

#define                 pATLAS_VOID                       0         // 00000000
#define                 pATLAS_WALL                       2         // 00000010
#define                 pATLAS_FREE                       164       // 10100100 palette number of DARK_GRAY
#define                 pATLAS_NOGO                       1         // 00000001
#define                 pATLAS_INJC                       8         // 00001000
#define                 pATLAS_MARK                       255       // 00001000

SDL_Surface*            pSnippet;
// END: globals

#include "pbrain.cpp"


void pInit() {
  int c1,c2;
  int cnt,pin,nei,tok;
  jDEBUG("pInit")  
  cFlag       =SDL_LoadBMP("res/cflag.bmp");         
  pFlag       =SDL_LoadBMP("res/pflag.bmp");         
  pAtlas      =SDL_LoadBMP("res/patlas.bmp"); // only right door to kitchen is opened
  //pAtlas      =SDL_LoadBMP("res/matlas.bmp"); // both doors to kitchen are opened
  pSneak      =SDL_LoadBMP("res/psneak.bmp");
  pSnippet    =SDL_CREATERGBSURFACE(230,173);
  if (cFlag) SDL_SetColorKey(cFlag,SDL_SRCCOLORKEY,SDL_MapRGB(cFlag->format,0,0,0));
  if (pFlag) SDL_SetColorKey(pFlag,SDL_SRCCOLORKEY,SDL_MapRGB(pFlag->format,0,0,0));
  if (pAtlas) {
    SDL_SetColorKey(pAtlas,SDL_SRCCOLORKEY,SDL_MapRGB(pAtlas->format,0,0,0));
    SDL_SetColorKey(pSneak,SDL_SRCCOLORKEY,SDL_MapRGB(pSneak->format,0,0,0));
    pTarget.data[0]=pLander.data[2*pLANDER_COFFEEWELL+1]*jSNK2GND;
    pTarget.data[1]=pLander.data[2*pLANDER_COFFEEWELL+0]*jSNK2GND;
    pTarget.data[2]=1;
    pAtlasConv();
  }
  if (pSneak) {
    SDL_FillRect(cTerrainStub,NULL,0);
    SDL_LockSurface(pSneak);
    SDL_LockSurface(cTerrainStub);
    Uint8* ptr=(Uint8*)(pSneak->pixels);
    sprintf(jChar,"SNEAK HEADER 0 %i 1 %i 2 %i",ptr[0],ptr[1],ptr[2]);
    dPrint(1,jChar);
    Uint32* myStub=(Uint32*)(cTerrainStub->pixels);
    pSneakNum=pSNEAK_SZE;
    for (c1=0;c1<cTERRAIN;c1++)
      if (ptr[c1]==pATLAS_FREE) {
        ptr[c1]=0;
        pSneakInt[pSneakNum+0]=0;
        pSneakInt[pSneakNum+1]=0;
        pSneakInt[pSneakNum+2]=0;
        pSneakInt[pSneakNum+pSNEAK_PRE]=0;
        pSneakInt[pSneakNum+pSNEAK_PIX]=c1%jMAP_W;
        pSneakInt[pSneakNum+pSNEAK_PIY]=c1/jMAP_W;
        pSneakInt[pSneakNum+pSNEAK_PIN]=c1;
        myStub[c1]=pSneakNum;
        pSneakNum+=pSNEAK_SZE;
      }
    cnt=0;
    for (c1=pSNEAK_SZE;c1<pSneakNum;c1+=pSNEAK_SZE) {
      pin=pSneakInt[c1+pSNEAK_PIN];
      tok=0;
      for (c2=0;c2<8;c2++) {
        nei=myStub[pin+jRouteOfs[c2]];
        if (nei&&tok<3) {
          pSneakInt[c1+tok]=nei;
          tok++;
        }
      }
      if (tok==3) {
        sprintf(jChar,"SNEAK TRIPLES %i x=%i y=%i",pin,pSneakInt[c1+pSNEAK_PIX],pSneakInt[c1+pSNEAK_PIY]);
        dPrint(1,jChar);
      }
    }
    SDL_UnlockSurface(cTerrainStub);
    SDL_UnlockSurface(pSneak);
        
    pLander.data[2*pLANDER_BIRTHPLACE  +0]=320;
    pLander.data[2*pLANDER_BIRTHPLACE  +1]=280;    
    pLander.data[2*pLANDER_CENTERALPARK+0]=220;
    pLander.data[2*pLANDER_CENTERALPARK+1]=390;    
    pLander.data[2*pLANDER_COFFEEWELL  +0]=325;
    pLander.data[2*pLANDER_COFFEEWELL  +1]=190;    
    pLander.data[2*pLANDER_ROBOTSHEAVEN+0]= 84;
    pLander.data[2*pLANDER_ROBOTSHEAVEN+1]=542;    
    pLander.data[2*pLANDER_TOXICWASTE  +0]=460;
    pLander.data[2*pLANDER_TOXICWASTE  +1]= 97;    
    pLanderLabel[pLANDER_BIRTHPLACE  ]=std::string("birth place");
    pLanderLabel[pLANDER_CENTERALPARK]=std::string("centeral park");
    pLanderLabel[pLANDER_COFFEEWELL  ]=std::string("coffee well");
    pLanderLabel[pLANDER_ROBOTSHEAVEN]=std::string("robots heaven");
    pLanderLabel[pLANDER_TOXICWASTE  ]=std::string("toxic waste");    
  }
  pTarget=jMover->pos.aff_pnt(2);
  pMoves.append(jMover);
  pMoves.append(mMover);
  double ang;
  pHuman.numel=0;
  for (c1=0;c1<pHuman.size[1];c1++) {
    ang=c1*360*jDEG2RAD/pHuman.size[1];
    pHuman.data[pHuman.numel++]=jHUMAN_RADIUS*cos(ang);
    pHuman.data[pHuman.numel++]=jHUMAN_RADIUS*sin(ang);
    pHuman.data[pHuman.numel++]=1;
  }
}

void pAddState(int myState) {
  pStates[pIndex++]=myState;
  pStates[pIndex  ]=pSTATE_IDLE;                      // just in case one forgets idle state
}

void pProcess(SDL_Event &mySDL_Event) {
  jDEBUG("pProcess");
  int c1,c2;
  if (mySDL_Event.type==SDL_KEYDOWN) {
    switch (mySDL_Event.key.keysym.sym) {
      case SDLK_0: // key zero terminates everything any other plans
        pSTATE_RESET;
        jSetSpeed*=0;
        mSetSpeed*=0;
        jTog[jTOG_IOPIN_LASER]=false;
        jTog[jTOG_IOPIN_LIGHT]=false;        
        break;
      case SDLK_1:
        pSTATE_RESET;
        pTargetList.clear();
        pTargetList.append(pTarget);
        pAddState(pSTATE_GUIDANCE);
        pIndex=0;        
        break;
      case SDLK_2:
        pSTATE_RESET;
        pTargetList.clear();
        pRoundtrip();
        pAddState(pSTATE_GUIDANCE);
        pIndex=0;    
        break;        
      case SDLK_3: pSTATE_RESET; pAddState(pSTATE_AFFECTION);             pIndex=0; break;
      case SDLK_4: pSTATE_RESET; pAddState(pSTATE_MBOT_INTERPLAY);        pIndex=0; break;
      case SDLK_5: pSTATE_RESET; pAddState(pSTATE_MBOT_CIRCLE);           pIndex=0; break;
      case SDLK_6: pSTATE_RESET; pAddState(pSTATE_MBOT_DEBUG); mJokers(); pIndex=0; break;
      case SDLK_8: pSTATE_RESET; pAddState(pSTATE_DRIVE_ALONG_PATH);      pIndex=0; break;
      case SDLK_r: {
        // BEGIN: clear trace and solids
        pTrace.size[1]=0; // reset counter
        pTrace.numel=0;
        c2=pTRACE*pTrace.size[0];
        for (c1=0;c1<c2;c1++)
          pTrace.data[c1]=0;
        cSolid.clear();
        // END: clear trace and solids
        jTog[jTOG_GLOBAL_RECORD]=!jTog[jTOG_GLOBAL_RECORD];
        break;
      }
      case SDLK_s: pRecordStore();                                   break;
  	}
  }
  if (mySDL_Event.type==SDL_MOUSEBUTTONDOWN) {
    int pix=mySDL_Event.button.x;
    int piy=mySDL_Event.button.y;
    if (jHUD_MAP_X<pix&&pix<jHUD_MAP_X+jMAP_W&&jHUD_MAP_Y<piy&&piy<jHUD_MAP_Y+jMAP_H)
      if (cDisplay!=cDISPLAY_ATLAS) {
        if (mySDL_Event.button.button==SDL_BUTTON_MIDDLE) {
          jTog[jTOG_MOUSE_MIDDLE]=true;
          mMouse.data[0]=pPos.data[6]+(piy-jHUD_MAP_Y-jMAP_Y)*jMAP2GND;
          mMouse.data[1]=pPos.data[7]+(pix-jHUD_MAP_X-jMAP_X)*jMAP2GND;
          mMouse.data[2]=1;
        }
        if (mySDL_Event.button.button==SDL_BUTTON_RIGHT) {
          cReset(false);
          jTog[jTOG_ATLAS_CALIBRATE]=true;
        }
      } else {
        if (mySDL_Event.button.button==SDL_BUTTON_LEFT) {
          pTarget.data[0]=(piy-jHUD_MAP_Y)*jSNK2GND;
          pTarget.data[1]=(pix-jHUD_MAP_X)*jSNK2GND;
          pTarget.data[2]=1;
        }
      }
  }
  if (mySDL_Event.type==SDL_MOUSEBUTTONUP) {
    if (mySDL_Event.button.button==SDL_BUTTON_RIGHT&&jTog[jTOG_ATLAS_CALIBRATE]) {
      cReset(false);
      jTog[jTOG_ATLAS_CALIBRATE ]=false;
      jTog[jTOG_ATLAS_CALIBRATED]=true;      
    }
    if (mySDL_Event.button.button==SDL_BUTTON_MIDDLE) {
      if (!jTog[jTOG_ALIGN_MBOT])
        pTarget=mMouse;
      jTog[jTOG_ALIGN_MBOT]=false;
      jTog[jTOG_MOUSE_MIDDLE]=false;
    }
  }
  if (mySDL_Event.type==SDL_MOUSEMOTION) {
    if (jTog[jTOG_ATLAS_CALIBRATE]) {
      pPos.data[6]-=mySDL_Event.motion.yrel*jMAP2GND;
      pPos.data[7]-=mySDL_Event.motion.xrel*jMAP2GND;      
      pInv.data[6]=-pPos.data[6];
      pInv.data[7]=-pPos.data[7];
    }
    if (jTog[jTOG_MOUSE_MIDDLE]) {
      matrix<float> dir(2,1);
      dir.data[0]=mMouse.data[0]-(pPos.data[6]+(mySDL_Event.motion.y-jHUD_MAP_Y-jMAP_Y)*jMAP2GND);
      dir.data[1]=mMouse.data[1]-(pPos.data[7]+(mySDL_Event.motion.x-jHUD_MAP_X-jMAP_X)*jMAP2GND);
      if (0.1<dir.norm(2)) {
        jTog[jTOG_ALIGN_MBOT]=true;
        dir.normalize(2);
        matrix<float> eye(3,3,"eye");
        eye.data[0]= dir.data[1];
        eye.data[1]=-dir.data[0];
        eye.data[3]= dir.data[0];
        eye.data[4]= dir.data[1];
        eye.data[6]=mMouse.data[0];
        eye.data[7]=mMouse.data[1];
        if (cDisplay==cDISPLAY_GEOGRAPHY) {
          dir*=mTEASER_OFFSET;
          eye.data[6]-=dir.data[0];
          eye.data[7]-=dir.data[1];
          mMover->pos=eye;
          mMover->track.clear();
          mMover->track.prepend(mMover->pos);
        }
        if (cDisplay==cDISPLAY_ROUTING)
          mTarget=eye;
      }
    }
  }
}

// state management
void pPlanStates(void) {
  jDEBUG("pPlanStates");
  pSTATE_RESET;
  if (jTic[jTOC_NEXT_RECHARGE]<jTic[jTOC_LAST_SENSORS]+jPOWER_TOLERANCE) {
    jTic[jTOC_NEXT_RECHARGE]=jTic[jTOC_LAST_SENSORS]+100000;            // 100 sec to drive back
    pTargetList.clear();    
    matrix<float> pos(3,1);
    pos.data[0]=pLander.data[2*pLANDER_BIRTHPLACE+1]*jSNK2GND;
    pos.data[1]=pLander.data[2*pLANDER_BIRTHPLACE+0]*jSNK2GND;
    pos.data[2]=1;
    pTargetList.append(pos);
    pAddState(pSTATE_GUIDANCE);
    pAddState(pSTATE_RECHARGE);
  } else
    switch (jBumperLast) {
      case jBUMPER_BACK: {
        switch (jBumperCount) {
          case jBUMPER_ONCE:                                            // pursuit
            if (jMover->stamp+jROLLOUT<jTic[jTOC_LAST_SENSORS]) {       // if stopped
              matrix<float> tmp=cPos.aff_pnt(2)-cStopped;
              if (pDIST_PURSUIT<tmp.norm(2)) {                          // robot has way to go
                tmp.normalize(2);
                pPursue=tmp;
                pAddState(pSTATE_PURSUIT);
              } else
                jSpeaks(jSPEAK_DRAGFURTHER);
            } else
              jSpeaks(jSPEAK_NOTSTOPPED);
            break;
          case jBUMPER_TWICE:
            pAddState(pSTATE_SELECT_TARGET);                            // puts one or no target in targetlist
            pAddState(pSTATE_GUIDANCE);
            break;                                                      // guiding to one
          case jBUMPER_THRICE:
            pTargetList.clear();
            pRoundtrip();
            pAddState(pSTATE_GUIDANCE);
            break;
        }
        break;
      }
      case jBUMPER_FRONT: {
        switch (jBumperCount) {
          case jBUMPER_ONCE:  pAddState(pSTATE_AFFECTION); break;
          case jBUMPER_TWICE:                                           // pursuit
            cReset(pStone); // why not in pRegister==0 !?
            pAddState(pSTATE_TRAIN_MARKER);
            break;
        }
        break;
      }
      case jBUMPER_LEFT: {
        switch (jBumperCount) {
          case jBUMPER_ONCE: pAddState(pSTATE_REMOTE_CONTROL); break;   // emulating joystick to control mbot
        }
        break;
      }
      case jBUMPER_RIGHT: {
        switch (jBumperCount) {
          case jBUMPER_ONCE:                                            // pursuit
            // put your code here
            break;
        }
        break;
      }
    }
  pIndex=0;
  jDrawString(screen,jFont,"!",7,pBAR_MASTER_Y);
}

// state control
void pControl(void) {
  jDEBUG("pControl");
  int c1,c2;
  bool next=false;
  int pState=pStates[pIndex];
  switch (pState) {
    //-----------------------------------------------------------------------------
    case pSTATE_IDLE:                               // NONE
      jSpeaks(jSPEAK_IDLE);
      jSpeaks(jSPEAK_RECHARGE);
      jTog[jTOG_BUMPER_COUNTING]=true;
      jTog[jTOG_LOOKING_FOR_HUMAN]=false;
      pPlanStates();
      mJokers();
      break;
    //-----------------------------------------------------------------------------
    case pSTATE_PURSUIT: {                          // SPEED, TITL
      if (pRegister==0) {
        jTog[jTOG_BUMPER_COUNTING]=false;
        jMover->active=1;
        mMover->active=0;
        pSTATE_TRY(4000);
        pRegister++;
      }
      if (pRegister==1) {
        matrix<float> prt=cLookInto(35,pPursue.data[0],pPursue.data[1]);
        matrix<float> com=cInv*prt;
        float ang=atan2(-com.data[0],com.data[1]);  // angle with respect to front, left is positive
        if (0.6f<prt.norm(2)) {                     // still way to go AND NOT bumper pressed
          jSpeaks(jSPEAK_WAY);
          jSetSpeed.data[0]=jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang);
          jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];
          jSetPTU.data[1]=jPTUTIL_PROFILE;
          pSTATE_TRY(4000);
        } else {
          jSpeaks(jSPEAK_NOWAY);
          jSetSpeed*=0;
          jSetPTU.data[1]=jPTUTIL_GESTURE;
          cROUTE_FREQUENT;                          // ensure that routing happens at a decent rate
        }
      }
      if (pSTATE_GIVEUP||jBumper) {
        jSpeaks(jSPEAK_FINISH);
        jSetSpeed*=0;
        jSetPTU.data[1]=jPTUTIL_NOOP;        
        next=true;
      }
      break;
    }
    //-----------------------------------------------------------------------------
    case pSTATE_SELECT_TARGET: {                      // PAN, TILT
      if (pRegister==0) {
        jSpeaks(jSPEAK_SELECT);
        jSetPTU.data[0]=0;
        jSetPTU.data[1]=jPTUTIL_MIN;                  // straight up
        jTog[jTOG_BUMPER_COUNTING]=true;
        pTargetList.clear();
        jDeclared.prepend(new std::string("[select init]"));
        pSTATE_TRY(20000);
        pRegister++;
      }
      if (pRegister==1) {
        if (jBUMPER_ONCE<=jBumperCount) {
          int ind=-1;
          matrix<float> tok(3,1);
          switch (jSingleton(jBumperLast)) {
            case jBUMPER_FRONT: ind=pLANDER_BIRTHPLACE;   break;
            case jBUMPER_LEFT:  ind=pLANDER_ROBOTSHEAVEN; break;
            case jBUMPER_BACK:  ind=pLANDER_CENTERALPARK; break;
            case jBUMPER_RIGHT: ind=pLANDER_COFFEEWELL;   break;
          }
          if (0<=ind) {
            jSpeak[jSPEAK_TARGET].appendix=&pLanderLabel[ind];            
            tok.data[0]=pLander.data[2*ind+1]*jSNK2GND;
            tok.data[1]=pLander.data[2*ind+0]*jSNK2GND;
            tok.data[2]=1;
            pTargetList.append(tok);            
            next=true;
          } else
            pRegister=0;
        }
      }
      if (pSTATE_GIVEUP) {
        jSpeaks(jSPEAK_GIVEUP);
        jSetPTU.data[1]=jPTUTIL_NOOP;
        next=true;
      }
      break;
    }
    //-----------------------------------------------------------------------------
    case pSTATE_TRAIN_MARKER: {                       // PAN, TILT
      jSetPTU.data[0]=0;
      jSetPTU.data[1]=jPTUTIL_MAX;
      jInt[jINT_MARKER_PIN]=0;                        // disable laser indication that marker is found
      jTog[jTOG_BUMPER_COUNTING]=false;               // any bumper press is good enough
      if (pRegister==0) {
        jSpeaks(jSPEAK_LASER);                
        jTog[jTOG_IOPIN_LASER]=true;
        jTog[jTOG_MARKER_CALIBRATION]=false;
        jPALETTE_EMPTY;
        pSTATE_TRY(20000);
        pRegister++;
      }
      if (pRegister==1) {
        if (jBumper) {
          jTog[jTOG_IOPIN_LASER]=false;
          jSpeaks(jSPEAK_SHUTTER);
          pSTATE_TRY(2000);                           // compensate delay of cam images
        }
        if (pSTATE_GIVEUP) {                          // => when no bumper +2000
          pSTATE_TRY(4000);                           // exposure time: 4000
          pRegister++;
        }
      }
      if (pRegister==2)                               // exposure
        if (camera) {
          jMarkerSet(183,117);
          jTog[jTOG_MARKER_CALIBRATION]=true;
        }
      if (pSTATE_GIVEUP||jCameraBrightness<jCAMERA_COVERED) {
        jTog[jTOG_IOPIN_LASER]=false;         // switch off laser
        jTog[jTOG_MARKER_CALIBRATION]=false;
        jPaletteUpdate();
        jSetPTU.data[0]=0;
        jSetPTU.data[1]=jPTUTIL_NOOP;
        next=true;
      }
      break;
    }    
    //-----------------------------------------------------------------------------
    case pSTATE_GUIDANCE: {                           // SPEED, PAN, TILT
      matrix<float> pos;
      jTog[jTOG_BUMPER_COUNTING]=false;               // graceful exit
      if (pRegister==0) {
        pMover=jMover;
        if (pTargetList.token(pos)) {
          pSTATE_TRY(6000);
          pTarget=pos;
          pTargetList.remove();
          pRegister++;
        } else {
          jSpeaks(jSPEAK_FINISH);
          next=true;
        }
      }
      if (pRegister==1) {
        jSpeaks(jSPEAK_TARGET);
        int pix,piy,pin;
        int Pix,Piy;        
        Pix=(int)(jMAP_X+cPos.data[7]*jGND2MAP);
        Piy=(int)(jMAP_Y+cPos.data[6]*jGND2MAP);                
        c1=0;
        while (pTrack.token(pos)&&c1<11) {            // about *8cm => 1.5 m
          pix=(int)(jMAP_X+(pos.data[1]-pPos.data[7])*jGND2MAP);
          piy=(int)(jMAP_Y+(pos.data[0]-pPos.data[6])*jGND2MAP);
          pin=piy*jMAP_W+pix;
          if (0<pin&&pin<cTERRAIN) {
            if (cRouteDist[pin]<cTERRAIN) {
              Pix=pix;
              Piy=piy;
              c1++;
            }
          } else
            break; // if the track point is out of the map the following points are not an option either
        }
        pTrack.next[2]=&pTrack;
        cTarget.data[0]=(Piy-jMAP_Y)*jMAP2GND;
        cTarget.data[1]=(Pix-jMAP_X)*jMAP2GND;
        matrix<float> com=cInv*jMover->plan.compass;
        float ang=atan2(-com.data[0],com.data[1]);    // angle with respect to front, left is positive
        if (jMover->plan.planned) {
          jSpeaks(jSPEAK_WAY);
          jSetSpeed.data[0]=jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang);
          jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];
          pSPEED_GOGETEMTIGER
          jSetPTU.data[1]=jPTUTIL_PROFILE;            // overwrite with oldptu when recording gesture
          pSTATE_TRY(4000);
        } else {                                      // no path available
          jSetSpeed*=0;
          jSpeaks(jSPEAK_NOWAY);
          jSetPTU.data[1]=jPTUTIL_NOOP;        
          cROUTE_FREQUENT;                            // ensure that routing happens at a decent rate
        }      
        matrix<float> tmp(2,1);
        tmp.data[0]=pPos.data[6]+cPos.data[6]-pTarget.data[0];
        tmp.data[1]=pPos.data[7]+cPos.data[7]-pTarget.data[1];
        if (tmp.norm(2)<0.3f) {                       // check if robot is close to goal
          jSpeaks(jSPEAK_ARRIVAL);
          pSTATE_TRY(-1000);
        }
      }
      if (pSTATE_GIVEUP||jBumper) {
        jSetSpeed*=0;
        jSetPTU.data[0]=0;
        jSetPTU.data[1]=jPTUTIL_NOOP;
        pRegister=0;
      }
      break;
    }
    //-----------------------------------------------------------------------------
    case pSTATE_AFFECTION: {                                          // SPEED, PAN, TILT    
      if (pRegister==0) {
        jTog[jTOG_LOOKING_FOR_HUMAN]=true;                              // declare pMover if no pMover and some human available
        jTog[jTOG_BUMPER_COUNTING]=false;                             // when bumper hit, exit affection state
        jMover->active=1; 
        mMover->active=0;
        pSTATE_TRY(12000);
        pRegister++;
      }
      jSetSpeed*=0;                                                   // dont move unless there is a reason            
      cTarget=cPos.aff_pnt(2);                                        // default for target is robot j2b2 itself
      if (pMover->id==jMOVITON_ID_HUMAN) {                            // BEGIN: robot has humanoid master
        pSTATE_TRY(12000);                                            // grant 20 seconds execution time
        matrix<float> pos=(cInv*pInv*pMover->pos).aff_vec(2);         // human in j2b2 coords
        float dst=pos.norm(2);
        sprintf(jChar,"has mover %4.2f %4.2f d=%6.3f last move %5.3f ID=%i HUMAN=%i",pos.data[0],pos.data[1],dst,(jTic[jTOC_LAST_SENSORS]-pMover->stamp)/1000.,pMover->id,jMOVITON_ID_HUMAN);
        jBulletin.append(new std::string(jChar));
        float hum=atan2(-pos.data[0],pos.data[1]);                    // angle for alignment (with respect to front, left is positive)
        if ((!pMOVER_VISUAL)||pDIST_FOLLOW<dst) {                     // BEGIN: robot does not have visual to human, or human is far away
          jBulletin.append(new std::string("not visual or distance"));
          if (jMover->plan.planned) {                                 // previous target was accessible by robot
            int pix,piy;
            matrix<float> yme;                                        // read out loud
            while (pMover->track.token(yme)) {                        // make earliest beacon of mover the new target
              pix=(int)(jMAP_X+(yme.data[7]-pPos.data[7])*jGND2MAP);
              piy=(int)(jMAP_Y+(yme.data[6]-pPos.data[6])*jGND2MAP);
              if (0<pix&&pix<jMAP_W&&0<piy&&piy<jMAP_H)
                if (cRouteDist[pix+piy*jMAP_W]<cTERRAIN) {
                  cTarget=pInv*yme.aff_pnt(2);
                  break;
                }
            }
            pMover->track.reset();                                    // reset track tokenizer          //pMover->track.next[2]=&pMover->track;                     
            if (0.4<jMover->plan.dist) {                              // follow path until default proximity  
              jBulletin.append(new std::string("distance greater than 0.4"));
              matrix<float> com=cInv*jMover->plan.compass;
              float ang=atan2(-com.data[0],com.data[1]);              // angle for driving (with respect to front, left is positive)
              jSetSpeed.data[0]=jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang);
              jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];
              pSPEED_GOGETEMTIGER
              if (!(pMOVER_VISUAL)) {                                 // mover not visible while approximating
                pMover->stamp=jTic[jTOC_LAST_SENSORS]-pMOVES_GRAVEYARD+1000; // artificially keep mover alive
                jBulletin.append(new std::string("artificially keep mover alive kill in 1 sec"));
              }
            }
          }
          if (pPTU_GRACEFUL) {
            jSetPTU.data[0]=-hum*1.2f;                                // look into target direction, MAGIC CONST
            if (pMOVER_VISUAL)                                        
              jSetPTU.data[1]=jRad2Til(atan(0.60f/MAX(0.30f,dst)));   // tilt cameras to look at mover
            else
              jSetPTU.data[1]=jPTUTIL_ABASHED;                        // indicate that human not visible
          }
        } else {                                                      // ELSE: human is visual and closer than pDIST_FOLLOW
          if (jTic[jTOC_STOP_ALIGNING]<jTic[jTOC_LAST_SENSORS]) {     // BEGIN: human is in vicinity of robot
            if (fabs(hum)<pDIST_WINDOW) {                             // human is in 35 field of view of robot
              matrix<float> prt=cLookInto(20,-cPos.data[3],-cPos.data[4]);
              if (dst<=pDIST_BACKUP&&0.55f<prt.norm(2)) {
                matrix<float> ret=cInv*prt;
                float ang=atan2( ret.data[0],-ret.data[1]);           // angle with respect to front, left is positive
                jSetSpeed.data[0]=-jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang);
                jSetSpeed.data[1]= pGainInt*ang-pGainPos*jVel.data[0];
                if (pPTU_GRANDMOTHER)
                  if (ang<0)
                    jSetPTU.data[0]=jPTUPAN_MIN;
                  else
                    jSetPTU.data[0]=jPTUPAN_MAX;
                jSetPTU.data[1]=jPTUTIL_ABASHED;                  
              } else 
                jLookAt(pos.data,jHUMAN_HEIGHT);                      // END: treat human in corridor                     
            } else 
              jTic[jTOC_STOP_ALIGNING]=jTic[jTOC_LAST_SENSORS]+3000;              
          } else {                                                    // align mode
            jSetSpeed.data[1]=pGainAng*hum-pGainOme*jVel.data[0];
            jLookAt(-hum*0.6f,jPTUTIL_ABASHED);
            if (20*jDEG2RAD<fabs(hum)||.1f<fabs(jVel.data[0]))
              jTic[jTOC_STOP_ALIGNING]=jTic[jTOC_LAST_SENSORS]+1500;
          }                                                           // END: human is in vicinity of robot
        }
      } else {                                                        // ELSE: robot does not have mover
        jSpeaks(jSPEAK_IDLE);
        jLookAround(jPTUTIL_NOOP);
      }
      if (pSTATE_GIVEUP||jBumper) {                                   // graceful exit
        jSpeaks(jSPEAK_FINISH);
        jSetSpeed*=0;
        jSetPTU.data[0]=0;
        jSetPTU.data[1]=jPTUTIL_NOOP;
        jTog[jTOG_LOOKING_FOR_HUMAN]=false;
        next=true;
      }
      break;
    }
    //-----------------------------------------------------------------------------
    case pSTATE_RECHARGE: {
      if (pRegister==0) {                                             // plug the plug and then press bumper
        jSpeaks(jSPEAK_POWERME);
        jSetPTU.data[1]=jPTUTIL_NOOP;
        jTog[jTOG_BUMPER_COUNTING]=false;                             // when bumper hit, next register
        if (jBumper)
          pRegister++;
      }
      if (pRegister==1&&!jBumper) {
        jTic[jTOC_NEXT_RECHARGE]=MAX(jTic[jTOC_LAST_SENSORS],jTic[jTOC_NEXT_RECHARGE]+1500);
        jSetPTU.data[0]=0;
        jSetPTU.data[1]=jPTUTIL_MIN;                                  // straight up
        pRegister++;
      }
      if (pRegister==2)
        if (jBumper) {
          jTic[jTOC_NEXT_RECHARGE]=jTic[jTOC_LAST_SENSORS]+jPOWER_DRIVE;
          jSetPTU.data[1]=jPTUTIL_NOOP;
          next=true;
        }
      break;
    }
    //-----------------------------------------------------------------------------    
    case pSTATE_REMOTE_CONTROL: { // rc of marsokhod
      mSetSpeed*=0;
      if (pRegister==0) {
        pSTATE_TRY(12000);
        pRegister++;
        jTic[jTOC_NEXT_LOOKAROUND]=jTic[jTOC_LAST_SENSORS]+2500;
        jTog[jTOG_JOYSTICK]=true;
      }              
      if (pRegister==1) {
        jSetPTU.data[1]=jPTUTIL_SAFETOWAVE;        
        if (jBumper&jBUMPER_FRONT) jAxis.data[0]=-mSPEED0_JOY;
        if (jBumper&jBUMPER_BACK)  jAxis.data[0]= mSPEED0_JOY;
        if (jBumper&jBUMPER_LEFT)  jAxis.data[1]= mSPEED1_JOY;
        if (jBumper&jBUMPER_RIGHT) jAxis.data[1]=-mSPEED1_JOY;
        if (jBumperLast==jBUMPER_BACK&&jBumperCount==jBUMPER_TWICE)
          jTog[jTOG_MBOT_ELONGATE]=!jTog[jTOG_MBOT_ELONGATE];
        if (jTog[jTOG_MBOT_ELONGATE]) 
          jButton.data[4]=jButton.data[5]=1;
        else
          jButton.data[6]=jButton.data[7]=1;
        if (jBumper)
          pSTATE_TRY(8000);
      }
      if (jTic[jTOC_NEXT_LOOKAROUND]<jTic[jTOC_LAST_SENSORS]) {
        jSetPTU.data[0]=0.8*(1-2*(0<jOldPTU.data[0]));
        jTic[jTOC_NEXT_LOOKAROUND]=jTic[jTOC_LAST_SENSORS]+2500;
      }
      if (pSTATE_GIVEUP) {
        jSetPTU.data[1]=jPTUTIL_NOOP;
        jTog[jTOG_JOYSTICK]=false;
        next=true;
      }
      break;
    }    
    //-----------------------------------------------------------------------------
    case pSTATE_MBOT_INTERPLAY: {
      mSetSpeed*=0;
      jSetSpeed*=0;
      if (pRegister==1) { // initially the targets are fresh
        if (mJokers(jTic[jTOC_NEXT_JOKER]<jTic[jTOC_LAST_SENSORS],mJOKER_AURORA_|mJOKER_WIGGLE_|mJOKER_TINYWIGGLE_)) {
          mLed|=mLED_WHITE;
          jTic[jTOC_NEXT_JOKER]=jTic[jTOC_LAST_SENSORS]+1500;
          pSTATE_TRY(3000)
        } else {
          int pix,piy,pin;
          int Pix,Piy;
          cTarget=mPos.aff_pnt(2);
          matrix<float> pos;
          c1=0;
          while (pTrack.token(pos)&&c1<17) {            // about *8cm => 1.2 m        
            pix=(int)(jMAP_X+(pos.data[1]-pPos.data[7])*jGND2MAP);
            piy=(int)(jMAP_Y+(pos.data[0]-pPos.data[6])*jGND2MAP);
            pin=piy*jMAP_W+pix;
            if (0<pin&&pin<cTERRAIN) {
              if (cRouteDist[pin]<cTERRAIN) {
                cTarget=pInv*pos;
                c1++;
              }
            } else
              break; // if the track point is out of the map the following points are not an option either
          }
          pTrack.next[2]=&pTrack; // reset tokenizer
          if (mTargets()) {
            pSTATE_TRY(3000)
            jTic[jTOC_NEXT_JOKER]=jTic[jTOC_LAST_SENSORS]+2000;
            mFollowed();
          }
          if (mMover->pos.solve(pTarget).aff_vec(0).norm(2)<0.25f) {
            jDeclared.prepend(new std::string("target reached"));
            pSTATE_TRY(-1000)
          }
        }
      }
      if (pRegister==0) { // initially request a lot of services:
        pSTATE_TRY(3000)
        jTog[jTOG_BUMPER_COUNTING]=false;
        jTog[jTOG_JOYSTICK]=false;
        cTarget=mPos.aff_pnt(2);
        pMover=mMover;                                        // tracking
        jMover->active=1;
        mMover->active=1;
        pRegister+=jTog[jTOG_FRESH_TRACKING];
        jTic[jTOC_NEXT_JOKER]=jTic[jTOC_LAST_SENSORS]+1500;
      }
      if (jBumper||pSTATE_GIVEUP) {
        jSetSpeed*=0;
        mSetSpeed*=0;
        jMover->active=1;        
        pMover=jMover;
        next=true;
      }
      break;
    }
    //-----------------------------------------------------------------------------
    case pSTATE_MBOT_DEBUG: {
      if (pRegister==0) {
        jTog[jTOG_BUMPER_COUNTING]=false;
        jTog[jTOG_JOYSTICK]=false;
        jTic[jTOC_STOP_MBOT_TARGET]  =jTic[jTOC_LAST_SENSORS]+60000;
        jTic[jTOC_STOP_MBOT_TOLERATE]=jTic[jTOC_LAST_SENSORS]+2000;
        jMover->active=1;
        mMover->active=1;
        mTarget=mMover->pos;
        mTarget.data[6]-=mTarget.data[3]*0.1f;
        mTarget.data[7]-=mTarget.data[4]*0.1f;
        mTarget.data[0]=-mTarget.data[0];
        mTarget.data[1]=-mTarget.data[1];
        mTarget.data[3]=-mTarget.data[3];
        mTarget.data[4]=-mTarget.data[4];
        mRouting();
        mMarking();
        sprintf(jChar,"marks %d",cCount(mMarks));        
        jDeclared.prepend(new std::string(jChar));
        if (cCount(mMarks)) 
          pRegister++;
        else {
          jDeclared.prepend(new std::string("turning somehow not possible"));
          next=true;
        }
      }
      if (pRegister==1) {
        sprintf(jChar,"stop mark in %6.3f sec %6.3f tolerate",
          (jTic[jTOC_STOP_MBOT_TARGET]-jTic[jTOC_LAST_SENSORS]  )*0.001f,
          (jTic[jTOC_STOP_MBOT_TOLERATE]-jTic[jTOC_LAST_SENSORS])*0.001f);
        jBulletin.append(new std::string(jChar));        
        next=!mTouring();
      }
      if (next||jBumper) {
        jSetSpeed*=0;
        mSetSpeed*=0;
        jMover->active=1;
        next=true;
      }
      break;
    }
    //-----------------------------------------------------------------------------
    case pSTATE_DRIVE_ALONG_PATH: {
      jSetSpeed*=0;
      jSetPTU.data[1]=jPTUTIL_PROFILE;
      jTog[jTOG_BUMPER_COUNTING]=false;
      if (jMover->plan.planned) {
        matrix<float> com=cInv*jMover->plan.compass;
        float ang=atan2(-com.data[0],com.data[1]);                    // angle with respect to front, left is positive
        if (0.3f<jMover->plan.dist) {                                        // end of plan is reached   cPlanDist cPlanDist cPlanDist !!!        
          jSetSpeed.data[0]=jSPEED0_SLW*MAX(0.0f,1.0f-pGainFwd*ang*ang);
          jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];
          pSPEED_GOGETEMTIGER
        }
      } else
        cROUTE_FREQUENT;                                              // ensure that routing happens at a decent rate
      if (next=jBumper)
        jSetSpeed*=0;
      break;
    }
    //-----------------------------------------------------------------------------    
    case pSTATE_DEBUG: {
      if (pRegister==0) {
        jSpeaks(jSPEAK_IDLE);
        jSetPTU.data[1]=jPTUTIL_GESTURE;        
        jTog[jTOG_BUMPER_COUNTING]=false;
        pSTATE_TRY(10000);
      }
      if (pSTATE_GIVEUP||jBumper) {
        jSetPTU.data[1]=jPTUTIL_NOOP;
        next=true;
      }
      break;
    }    
  }
  int cnt=0;
  jDrawString(screen,jFont,"s",0,pBAR_MASTER_Y);
  while (pStates[cnt]) {
    if (cnt==pIndex)
      sprintf(jChar,">%i",pStates[cnt]);
    else
      sprintf(jChar," %i",pStates[cnt]);
    cnt++;
    jDrawString(screen,jFont,jChar,cnt*21+7,pBAR_MASTER_Y);
  }
  sprintf(jChar,"r%i",pRegister);
  jDrawString(screen,jFont,jChar,pIndex*21+7,pBAR_MASTER_Y+11);  
  if (next) {
    pIndex++;
    jTic[jTOC_NEXT_STATE]=0;
    pRegister=0;
  }
  if (!(pSTATE_GIVEUP)) {
    sprintf(jChar,"time to giveup %5.1f",(jTic[jTOC_NEXT_STATE]-jTic[jTOC_LAST_SENSORS])/(float)1000);
    jBulletin.append(new std::string(jChar));
  }
}

void jDrawBot(char* str,int rad,float* data) {
  int pix,piy;
  pix=(int)(data[7]*jGND2SNK);
  piy=(int)(data[6]*jGND2SNK);
  ellipseRGBA(cMap,pix,piy,rad,rad,0x00,0x00,0x80,0xff);
  ellipseRGBA(cMap,pix,piy,rad,rad,0x00,0x00,0x80,0xff);
  jDrawString(cMap,jFong,str,pix-3,piy-4);
}

void pHUD(void) {
  jDEBUG("pHUD")
  int c1,c2;
  int pix,piy,pin;
  int rad;
  SDL_Rect mySDL_Rect;
  SDL_Rect mySDL_src;
  SDL_Rect mySDL_dst;
	jMoviton* tok;
	if (mJoker!=mJOKER) {
	  sprintf(jChar,"The Joker <%d> reg %d",mJoker,mJokerReg);
	  jBulletin.append(new std::string(jChar));
  }
	
  if (cDisplay!=cDISPLAY_ATLAS) {
    // BEGIN: draw track planned in sneak
    jDraw(jIN_MAP,(pInv*pTarget).data,2,SDL_MapRGBA(&jSDL_PixelFormat, 64,255,255,0xff),&filledEllipseRGBA);
    jDraw(jIN_MAP,(pInv*pTarget).data,2,SDL_MapRGBA(&jSDL_PixelFormat, 64,196,196,0xff),      &ellipseRGBA);
    matrix<float> pos(3,1);
    while (pTrack.token(pos))
      if (jAtlas2map(pos,pix,piy))
      pixelRGBA(cMap,pix,piy,224,224,224,255);    
    // BEGIN: display humans
    Uint32 col;
    matrix<float> tri(3,3);
    tri.data[0]=0; tri.data[1]=jAXIS_OFFSET+jROBOT_RADIUS; tri.data[2]=1;
    tri.data[3]=-.14; tri.data[4]=-.14; tri.data[5]=1;
    tri.data[6]=+.14; tri.data[7]=-.14; tri.data[8]=1;          
    while (pMoves.token(tok)) {
	    //sprintf(jChar,"moves id= %i (%i)",tok->id,tok==pMover);
      //jBulletin.append(new std::string(jChar));
      col=cMOVER_COLOR(tok);
      if (jMOVITON_ID_HUMAN<tok->id)
        jDrawPlan(tok->plan);
      jDraw(jIN_MAP,pInv*tok->pos**tok->contour,cMOVER_COLOR(tok),&polygonRGBA);
      switch (cDisplay) {
        case cDISPLAY_GEOGRAPHY: {          
          if (tok->id) 
            jDraw(jIN_MAP,pInv*tok->pos*tri,pMOVER_COLOR(tok),&polygonRGBA);
          if (tok->visual!=jTic[jTOC_LAST_SENSORS]&&jAtlas2map(tok->pos.aff_pnt(2),pix,piy))
            jDrawString(cMap,jFonr,"v",pix,piy);
          break;
        }
        case cDISPLAY_ROUTING: {
          jDrawFrame(jIN_MAP,pInv*tok->pos,0xff000000);
          if (tok->id==jMOVITON_ID_HUMAN)
            jDraw(jIN_MAP,pMovesAura(tok->pos)                  ,col,&polygonRGBA);      
          int pjx,pjy;
          matrix<float> yme;
          if (jAtlas2map(tok->pos.aff_pnt(2),pix,piy)) {
            c1=0;
            while (tok->track.token(yme)&&c1<30) {
              jAtlas2map(yme.aff_pnt(2),pjx,pjy);
              lineRGBA(cMap,pix,piy,pjx,pjy,196,128,64,255);
              pix=pjx;
              piy=pjy;
              c1++;
              if (0<pix&&pix<jMAP_W&&0<piy&&piy<jMAP_H)
                if (cRouteDist[pix+piy*jMAP_W]<cTERRAIN)
                  ellipseRGBA(cMap,pix,piy,2,2,128,64,32,255);
            }
            tok->track.reset();
          }
        }
      }
    }
    matrix<float> ofs(3,1);
    // BEGIN: robot
    for (c1=0;c1<bumper.distanceArray.size();c1++)
      if (0<bumper.distanceArray[c1].distance) {
        ofs.data[0]=-sin(c1*90*jDEG2RAD)*jROBOT_RADIUS;
        ofs.data[1]= cos(c1*90*jDEG2RAD)*jROBOT_RADIUS+jAXIS_OFFSET;
        ofs.data[2]= 1;
        jDraw(jIN_MAP,(cPos*ofs).data,1,0xffff0000,&filledEllipseRGBA);
      }
    // END: robot
    // BEGIN: marsokhod
    jDrawFrame(jIN_MAP,pInv*mTarget,0xff0000ff);
    mMarkingDraw();
    col=cMOVER_COLOR(mMover);
    jDraw(jIN_MAP,(mPos*mOfs).data,(int)(mTEASER_RADIUS*jGND2MAP),col,&ellipseRGBA);
    for (c1=mMOBILE_FRONT;c1<=mMOBILE_RIGHT;c1++)
      if (!mMobility.data[c1]) {
        ofs.data[0]= sin(c1*90*jDEG2RAD)*mSIZE0;
        ofs.data[1]= cos(c1*90*jDEG2RAD)*mSIZE1;
        ofs.data[2]= 1;
        jDraw(jIN_MAP,(mPos*ofs).data,1,0xffff0000,&filledEllipseRGBA);
      }    
    // END: marsokhod    
    
    // END: display humans
    while (pStone.token(tok)) {      
      mySDL_Rect.x=(int)(jMAP_X+(tok->pos.data[7]-pPos.data[7])*jGND2MAP-4);
      mySDL_Rect.y=(int)(jMAP_Y+(tok->pos.data[6]-pPos.data[6])*jGND2MAP-13);
      SDL_BlitSurface(cFlag,NULL,cMap,&mySDL_Rect);      
    }
    // END: draw detected motion
    mySDL_Rect.x=jHUD_MAP_X;
	  mySDL_Rect.y=jHUD_MAP_Y;
    if (jTog[jTOG_ATLAS_CALIBRATE]) {
      SDL_Rect atl;
      atl.x=(int)(jHUD_MAP_X+jMAP_X-pPos.data[7]*jGND2MAP);
      atl.y=(int)(jHUD_MAP_Y+jMAP_Y-pPos.data[6]*jGND2MAP);
      SDL_BlitSurface(pAtlas,NULL,screen,&atl);
	  }
    // END: draw track planned in sneak    
    for (c1=0;c1<mJOKER;c1++) {
      pix=(int)(jMAP_X+(mJokerPos[c1].data[7]-pPos.data[7])*jGND2MAP);
      piy=(int)(jMAP_Y+(mJokerPos[c1].data[6]-pPos.data[6])*jGND2MAP);
      if (0<pix&&pix<jMAP_W&&0<piy&&piy<jMAP_H) {
        sprintf(jChar,"j%d",c1);
        jDrawString(cMap,jFont,jChar,pix,piy);
      }
    }
  } else {  // DISPLAY_ATLAS
    mySDL_Rect.x=jHUD_MAP_X;
	  mySDL_Rect.y=jHUD_MAP_Y;
    SDL_BlitSurface(pSneak,NULL,screen,&mySDL_Rect);
    mySDL_Rect.x=jMouse.data[0]-jHUD_MAP_X;
    mySDL_Rect.y=jMouse.data[1]-jHUD_MAP_Y;
    if (0<mySDL_Rect.x&&0<mySDL_Rect.y) {
	    sprintf(jChar,"mouse at (%6.2f,%6.2f)",mySDL_Rect.y*jSNK2GND,mySDL_Rect.x*jSNK2GND);
      jBulletin.append(new std::string(jChar));
      // BEGIN: draw landmarks
      for (c1=0;c1<pLANDER;c1++) {
        filledEllipseRGBA(cMap,pLander.data[c1*2+0],pLander.data[c1*2+1],3,3,0xff,0x00,0xff,0xff);
              ellipseRGBA(cMap,pLander.data[c1*2+0],pLander.data[c1*2+1],3,3,0x00,0x00,0x00,0xff);
        jDrawString(cMap,jFong,pLanderLabel[c1].c_str(),pLander.data[c1*2+0]+4,pLander.data[c1*2+1]+2);
      }
      // END: draw landmarks    
    }
    // BEGIN: draw global track
    matrix<float> tps;
    while (pTrack.token(tps))
      ellipseRGBA(cMap,(int)(tps.data[1]*jGND2SNK),(int)(tps.data[0]*jGND2SNK),0,0,0x80,0x80,0x80,0xff);
    // END: draw global track
    // BEGIN: draw flags for detected marker places
    while (pStone.token(tok)) {
      mySDL_Rect.x=(int)(tok->pos.data[7]*jGND2SNK-1);
      mySDL_Rect.y=(int)(tok->pos.data[6]*jGND2SNK-6);
      SDL_BlitSurface(pFlag,NULL,cMap,&mySDL_Rect);
    }
    // END: draw flags for detected marker places
    // BEGIN: draw current position and final target
    rad=jTic[jTOC_LAST_SENSORS]/150;
    rad=10-(rad%10);
    jDrawBot("J",rad,jMover->pos.data);
    jDrawBot("M",rad,mMover->pos.data);
    ellipseRGBA(cMap,(int)(jGND2SNK*pTarget.data[1]),(int)(jGND2SNK*pTarget.data[0]),rad,rad,0x00,0x80,0x00,0xff);
    // END: draw current position and final target
  }
}
