#ifndef _SET_HPP_
#define _SET_HPP_

// <type> has to have an == operator overloaded

// set implements a hashSet as on
// http://java.sun.com/j2se/1.5.0/docs/api/java/util/HashSet.html
template <class type> class jset {
	public:
		jset<type>(void);
		jset<type>(int n,int (*f)(type));    // init with # of buckets and set function (ptr)
		~jset<type>(void);                   // destruct all
		bool add(const type &data);         // stores data
		void clear(void);        
		bool contains(const type &data);    // stored status of data
		bool remove(const type &data);      // removes data
		bool apply(bool (*f)(type));
		array<linked<type>*> table;
		int numel;													// cardinality of set
	private:
		int (*func)(type);
		type ref;
};
//constructors-------------------------------------------------------------
template <class type> jset<type>::jset(void) {
  numel=0;
  table=array<linked<type>*>(1,&numel,NULL);
};

template <class type> jset<type>::jset(int n,int (*f)(type)) {
  numel=0;
	table=array<linked<type>*>(1,&n,NULL);
	for (int c1=0;c1<table.numel;c1++)
		table.data[c1]=new linked<type>();
	func=f;	
};

template <class type> jset<type>::~jset(void) {
	for (int c1=0;c1<table.numel;c1++)
	  delete table.data[c1];
}

template <class type> bool jset<type>::add(const type &data) {
	int n=func(data)%table.numel;
	while (table.data[n]->token(ref))
		if (ref==data) {
			table.data[n]->reset();
			return false;
		}
	table.data[n]->append(data);
	numel++;
	return true;
}

template <class type> void jset<type>::clear(void) {
  for (int c1=0;c1<table.numel;c1++)
	  table.data[c1]->clear();
	numel=0;
}

template <class type> bool jset<type>::contains(const type &data) {
	int n=func(data)%table.numel;
	while (table.data[n]->token(ref))
		if (ref==data) {
			table.data[n]->reset();
			return true;
		}
	return false;
}

//this function can be called within another linked list tokenizer:-)
//behaviour is undefined when current gets removed.
//current can be removed from outside with
//while (set->...->token(bucket)) set->table[c1]->remove();
template <class type> bool jset<type>::remove(const type &data) {
	int n=func(data)%table.numel;	
	ntree<type> *pre=table.data[n]->next[2]; //due to this
	table.data[n]->reset();
	while (table.data[n]->token(ref))
		if (ref==data) {
			table.data[n]->remove();
			table.data[n]->reset();
			numel--;
			table.data[n]->next[2]=pre; //this
			return true;
		}
	table.data[n]->next[2]=pre; //and this
	return false;
}

template <class type> bool jset<type>::apply(bool (*f)(type)) {
	for (int c1=0;c1<table.numel;c1++)
		while (table.data[c1]->token(ref))
			if (!f(ref)) {
				table.data[c1]->reset();
				return false;
			}
	return true;
}

#endif
