// code by jph

import java.io.*;
import java.util.*;

public class Control {
  static double clk=0;
  static double upd=0;
  static double dlt=0.001;
  
  double    CN;         // IN max speed (mu is compensated for automatically)
  double    MU;         // IN parameter
  double    EP;         // IN max error for no action
  double    EV;         // IN max error for no action
  double    vd=0;       // IN velocity desired
  double    VD=0;       // IN velocity desired, entered by application software
  double    pd=0;       // IN position desired (not used in velocity control)
  double    vn=0;       // IN/OUT velocity
  double    pn=0;       // IN/OUT position
  double    ct=0;       // internal parameter
  double    dd=0;       // OUT intention of direction 0 [0<vd] or 1 [vd<0] (dd is preserved even if vd=0)
  double    cn=0;       // OUT control
  double    ei;         // OPTIONAL 
  
  public Control(double CN,double MU,double EP,double EV) {
    this.CN=CN;
    this.MU=MU;
    this.EP=EP;
    this.EV=EV;
  }
  
  private void integrate() {  // discretized exact equations
    double etm=Math.exp(-dlt*MU);
    pn=pn+vn*(1-etm)/MU-cn*(1-dlt*MU-etm)/MU;
    vn=vn*etm+cn*(1-etm);                   
    if (0<vd) dd=0;
    if (vd<0) dd=1;
  }
  
  public void iVD() {
    if (clk<upd)
      vd=VD;      // application software rules
    else
      vd=0;       // dead mans switch
  }
  
  private double cV(double vn,double vd) {
    return clip(-CN, (vd-vn*(1-MU*dlt))/(dlt*MU) ,CN);
  }
  
  private double cVP(double vn,double vd,double pn,double pd) {
    if (EP<Math.abs(pd-pn)||EV<Math.abs(vd-vn)) {
      ct=CN*sign(vd-vn);
      return CN*sign(pd-(pn+(vn-vd+ct*Math.log((ct-vn)/(ct-vd)))/MU));
    } else
      return cV(vn,vd);
  }
  
  public void V() {
    V(Double.NEGATIVE_INFINITY,Double.POSITIVE_INFINITY);
  }
  
  public void V(double plo,double phi) { // velocity control with bound lo and hi
    cn=clip(cVP(vn,0,pn,plo),cV(vn,vd),cVP(vn,0,pn,phi));
    integrate();
  }
  
  public void VP() {  // position and velocity control    
    cn=cVP(vn,vd,pn,pd);
    integrate();
  }
  
  public double VPerror() {
    if (EP<Math.abs(pd-pn)||EV<Math.abs(vd-vn))
      return 1;
    else
      return 0;
  }
  
  public boolean Perror() {
    return EP<Math.abs(pd-pn);
  }
  
  
  public boolean bP(double plo,double phi) {
    return plo-EP<pn&&pn<phi+EP;
  }
  
  
  public static double sign(double num) { // binary signature
    if (num<0) return -1; else return 1;
  }
  
  public static double sign(double num,double tol) { // binary signature
    if (num+tol<0)
      return -1;
    if (0<num-tol)
      return 1;
    else
      return 0;
  }

  public static double clip(double min,double num,double max) { // ensures number is within bounds
    return Math.min(Math.max(min,num),max);
  }

  public static double fmod(double num,double bas) { // remainder of division
    return num-bas*Math.floor(num/bas);
  }
  
  public static double sumabs(double[] vec) {
    int c1;
    double sum=0;
    for (c1=0;c1<vec.length;c1++)
      sum+=Math.abs(vec[c1]);
    return sum;
  }
  
}
