// code by jph

import java.io.*;
import java.net.*;

public class Mediator {
  static final int            TC=49401;                 // port of mediator.c
  static volatile String      IP="130.233.120.187";     // ip of mediator.c (via lan OR else wireless!)
  
  static final int[]          SGN={ 1,-1, 1,-1, 1,-1, 1,-1};   // velocity convention  
  static volatile long        ido=0;                    // difference between send and recv packages
  static final double         EIM=0.60;                 // beyond EIM<|duty| external interupt is deactivate
  static private boolean[]    eim=new boolean[8];       // totally handled by marsokhod:extended
  static volatile int[]       dir=new int[8];           // 0 or 1, notices direction change
    
  static Socket               mySocket;
  static DataOutputStream     myDataOutputStream;
  static DataInputStream      myDataInputStream;
  
  // this could be a simulator in here :-)
  //static public void main(String[] args) {
  //  try {
  //    ServerSocket myServerSocket=new ServerSocket(TC);
  //    while (true) {
  //      mySocket=myServerSocket.accept();
  //      myDataInputStream=new DataInputStream(mySocket.getInputStream());
  //      myDataOutputStream=new DataOutputStream(mySocket.getOutputStream());
  //      System.out.println(mySocket);
  //    }
  //  } catch (Exception myException) {
  //    System.out.println(myException);
  //  }
  //}
  
	static public void open() throws IOException {
  	System.out.print("mediator at "+IP+":"+TC+"... ");
    mySocket=new Socket(IP,TC);
    myDataOutputStream=new DataOutputStream(mySocket.getOutputStream());
    myDataInputStream=new DataInputStream(mySocket.getInputStream());
    System.out.println("ok.");
	}
	
  static public void close() throws IOException {
    mySocket.close();
	}
	
	static private long toc=0;
	
	static public synchronized void send() throws IOException {
  	int c1,c2;
  	int tmp;
  	byte[] send=new byte[34];
  	send[0]=(byte)0xf0;
  	for (c1=0;c1<8;c1++) {
    	c2=c1*3+1;
    	send[c2+0]=0x01;
      eim[c1]=Math.abs(Marsokhod.pwm[c1])<EIM*Marsokhod.icr[c1];
    	if (!eim[c1])
    	  send[c2+0] =0x00;
  	  if (Marsokhod.pwm[c1]<0)
  	    send[c2+0]|=0x04;
  	  tmp=Math.abs(Marsokhod.pwm[c1]);
  		send[c2+1]=(byte)(tmp>>8);
  		send[c2+2]=(byte) tmp;
		}
		String.format("%1d  %5.1f ",toc++%2,Marsokhod.SLA()*57.2958).getBytes(0,8,send,1+8*3);
		send[1+8*3+0]|=(Marsokhod.led&Marsokhod.LED_GREEN)<<7;
		send[1+8*3+1]|=(Marsokhod.led&Marsokhod.LED_WHITE)<<6|(Marsokhod.led&Marsokhod.LED_CALIB)<<5;
	  myDataOutputStream.write(send,0,33); // 1+8*3+8
 	  myDataOutputStream.flush();
 	  ++ido;
	}
	
	static public synchronized void sendCfg() throws IOException {
  	int c1,c2;
  	int tmp;
  	byte[] send=new byte[42];
  	send[0]=(byte)0xf2;
  	for (c1=0;c1<8;c1++) {
    	c2=1+c1*5;
  	  tmp=Math.abs(Marsokhod.pwm[c1]);
  	  send[c2+0]=(byte)(Marsokhod.icr[c1]>>8);
  	  send[c2+1]=(byte) Marsokhod.icr[c1] ;
  	  send[c2+2]=(byte)(tmp>>8);
  	  send[c2+3]=(byte) tmp ;
  	  send[c2+4]=(byte) Marsokhod.con[c1];
		}
  	myDataOutputStream.write(send,0,41);
  	myDataOutputStream.flush();
  	++ido;
  }
	
  static private int[]      buf_tic=new int[8];
  static private int[]      buf_enc=new int[8];
	
  static public synchronized void read() throws IOException {
  	int c1,p1;
  	int tmp,sum,slp=0;
  	int tic,enc,stp,vel,mdr;
  	double[] tmp_grv=new double[3];
  	double elp=0;
  	byte[] recv=new byte[128];
  	int[] read=new int[128];
  	String myString;
  	while (77<=myDataInputStream.available()) {
    	myDataInputStream.read(recv,0,77);
    	for (c1=0;c1<77;c1++) {
      	read[c1]=recv[c1];
        if (read[c1]<0)
          read[c1]+=256;
      }
      myString="";
    	switch (read[0]) {
      	case 0xf1:
      	  --ido;
      	  Marsokhod.msk=(read[1]<<8)+read[2];
      	  slp+=(read[3]<<8)+read[4];
      	  for (c1=0;c1<8;c1++)
      	    if ((Marsokhod.msk&1<<c1)!=0) {
        	    p1=5+c1*8;
    	        // BEGIN: parse obtained message
    	        tic=(read[p1+0]<<8)+read[p1+1];
    	        tmp=(tic+65536-buf_tic[c1])%65536;      	        
              if (62000<tmp)
                myString+=" "+c1+" "+tmp;
    	        elp+=tmp*Marsokhod.CONsec(Marsokhod.con[c1]);
    	        buf_tic[c1]=tic;
    	        enc=(read[p1+2]<<8)+read[p1+3];
    	        stp=(read[p1+4]<<8)+read[p1+5];
    	        vel=(read[p1+6]<<8)+read[p1+7]&0x7fff;
    	        mdr=(read[p1+6]&0x80)>>7;               // 0 or 1
      	      if (SGN[c1]<0)
        	      mdr=1-mdr;
    	        // END: parse obtained message
    	        if (!eim[c1])                           // update direction only if clock conv chip trustworthy              
                mdr=dir[c1];      	    
    	        if (mdr!=dir[c1]) {
      	        sum =(stp+65536-buf_enc[c1])%65536;   // add old ticks from NUM_CNT to STOP
      	        sum-=(enc+65536-stp        )%65536;   // subtract tics in the other direction:         	    
    	        } else
      	        sum =(enc+65536-buf_enc[c1])%65536;   // assumes a certain update rate
              
    	        if (0==dir[c1]) {
      	        Marsokhod.enc[c1]+= sum;
              } else {
      	        Marsokhod.enc[c1]-= sum;
      	        vel =-vel;
              }
              buf_enc[c1]=enc;
              Marsokhod.pos[c1]=Marsokhod.ENCrad(Marsokhod.enc[c1]);
              Marsokhod.vel[c1]=vel*0.00166737/(Marsokhod.con[c1]+1);    // 30.517578125/18302.81845
              dir[c1]=mdr;
    	      }
      	  c1=8;
      	    if ((Marsokhod.msk&1<<c1)!=0) {
        	    p1=5+c1*8;
        	    Marsokhod.yaw=(read[p1+0]<<2)+((read[p1+6]>>6)&0x03);
        	    Marsokhod.gyr=Marsokhod.ADCgyr(Marsokhod.yaw);
        	    Marsokhod.acc[0]=(read[p1+1]<<2)+((read[p1+6]>>4)&0x03);
        	    Marsokhod.acc[1]=(read[p1+2]<<2)+((read[p1+6]>>2)&0x03);
        	    Marsokhod.acc[2]=(read[p1+3]<<2)+((read[p1+6]>>0)&0x03);
        	    Marsokhod.ACCdigest();          	    
        	    tmp=read[p1+7]&0x01;
        	    Marsokhod.tlt[tmp]=Marsokhod.ADCrad(tmp,(read[p1+4]<<2)+((read[p1+7]>>6)&0x03));
        	    Marsokhod.vcc     =Marsokhod.ADCvcc(    (read[p1+5]<<2)+((read[p1+7]>>4)&0x03));
            }
      	  break;
      	case 0xf3:
      	  --ido;
      	  Marsokhod.msk=read[1];
      	  slp+=(read[2]<<8)+read[3];
      	  for (c1=0;c1<8;c1++)
      	    if ((Marsokhod.msk&1<<c1)!=0) {
        	    p1=4+c1*8;
    	        // BEGIN: parse obtained message
    	        tic=(read[p1+0]<<8)+read[p1+1];
    	        tmp=(tic+65536-buf_tic[c1])%65536;      	        
              if (62000<tmp)
                myString+=" "+c1+" "+tmp;                  
    	        elp+=tmp*Marsokhod.CONsec(Marsokhod.con[c1]);
    	        buf_tic[c1]=tic;      	        
    	      }
      	  break;
      	default:
      	  System.out.println(" mediator.c -> mediator.java :: corrupted tcp");
      	  break;
    	}
    	if (0<myString.length())
    	  System.out.println("time"+myString);      	
	  }
    elp*=0.125; // assuming eight wheels
    Control.clk+=elp;
    if (0<elp) {
      double lam=0.35;
      Control.dlt=Control.clip(0.015,Control.dlt*(1-lam)+elp*lam,0.032);      
    }
	}

  
}
