// code by jph based on code by amaula
// using GIM-Interface Version Dec-2007 <-> Aug-2008
// revised on 29-Aug-2008

#include "j2b2GIMICtrl.hpp"
#include "gimi.h"
#include "owndebug.h"

#define GIMI_CLIENT_API 10403
#define GIMI_TTS_SERVICE_ID 118

#include <SDL/SDL.h>
#include <SDL/SDL_gfxPrimitives.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_thread.h>

using namespace J2B2GIMICtrl;
static std::string own_group_id="real";

#define jDEBUG(x)                                     // no debugging
//#define jDEBUG(x) dPrint(1,x);                        // keep track of functions enter and exit

SDL_Surface* screen=NULL;
SDL_Surface* camera;            // only one camera, because we are not using stereo vision
ASRobo::Drivers::Devices::Ranging::CDistanceArray bumper;
ASRobo::Drivers::Devices::Ranging::CDistanceArray laser;

CJ2B2GIMICtrl ctrl;
gimi::GIMI    cgim(1);          // < 100 sized queue

#include <math.h>
#include "inc/ntree.hpp"
#include "inc/array.hpp"
#include "inc/matrix.hpp"
#include "inc/set.hpp"
#include "inc/hash.hpp"
#include "inc/jdraw.h"
#include "inc/jcode.cpp"
#include "inc/mcode.cpp"
#include "inc/ccode.cpp"
#include "inc/pcode.cpp"
#include "inc/jdraw.c"

int main(int argc, char *argv[]) {
  int c1;
  debugInit();
  debugSetLogFilename("j2b2gimictrl_example.log");
  debugSetPrintLevel(3);
  debugSetLogLevel(3);
  
  if (!ctrl.Connect("130.233.120.181",51002,"j2b2_jassnsback"+own_group_id)) {
    dPrint(1,"Failed to Connect J2B2GIMICtrl. Fatal error - bailing out.");
    exit(1);
  }
  
  if (!cgim.connectToHub("130.233.120.181", 51002))
    dPrint(1,"Failed to Initialize gim!");
  
  
  /*
  if (!ctrl.Connect("asrobo.hut.fi",50002,"j2b2somejuldea_"+own_group_id)) {
    dPrint(1,"Failed to Connect J2B2GIMICtrl. Fatal error - bailing out.");
    exit(1);
  }
  
  if (!cgim.connectToHub("asrobo.hut.fi",50002))
    dPrint(1,"Failed to Initialize gim.");
  */
  
  /*
  if (!ctrl.Connect("turbo.dyndns.tv",51002,"j2b2unabcissssquectrl_"+own_group_id)) {
    dPrint(1,"Failed to Connect J2B2GIMICtrl. Fatal error - bailing out.");
    exit(1);
  }
  
  if (!cgim.connectToHub("turbo.dyndns.tv",51002))
    dPrint(1,"Failed to Initialize gim.");
  */
  
  if (!ctrl.InitializeASRobo("ASRobo_"+own_group_id)) {
    dPrint(1,"Failed to Initialize ASRobo!!! Operating without it then.");
    ownSleep_ms(2000);
    exit(1);
  }  
  
  if (!ctrl.InitializePTU("PTU_"+own_group_id)) {
    dPrint(1,"Warning: Failed to Initialize PTU!!! Operating without it then.");
    ownSleep_ms(2000);
    exit(1);
  }
  
  if (!ctrl.InitializeImageClient("ImageServer_"+own_group_id)) {
    dPrint(1,"Warning: Failed to Initialize ImageClient!!! Operating without it then.");
    ownSleep_ms(2000);
    exit(1);
  }
  
  if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK)==-1) {
    dPrint(1,"Failed to initialize SDL (Video & Joystick): %s\n",SDL_GetError());
    exit(1);
  }
  
  atexit(SDL_Quit);
  
  screen=SDL_SetVideoMode(jSCREEN_W,jSCREEN_H,24,SDL_SWSURFACE|SDL_DOUBLEBUF);  
  SDL_WM_SetIcon(SDL_LoadBMP("res/jicon.bmp"),NULL);
  SDL_WM_SetCaption("J2B2 Encore",NULL);
  dPrint(1,"ASRobo - Drive - SetAcceleration to 0.2m/s");
  
  ctrl.ASRobo_SetAcceleration(0.2);         // acc always at 0.2m/s
  ctrl.ASRobo_SetAllBehaviours(false);  
  
  jInit();
  mInit();
  cInit();
  pInit();                                  // read atlas and psneak
  
  jRandu.size[1]=jCLOCK_FUNC%jRandu.numel;  
	
  SDL_Event event;
  int rch=0;
  matrix<float> bog(3,1);
  while (jTic[jTOC_LAST_SENSORS]<jTic[jTOC_NEXT_SHUTDOWN]+jSHUTDOWN_DELAY) {
      // BEGIN: perception and actuation
    if (jSensors()) {       // state matrix, camera precomputations      
      jGoggles();
      jLasers();            // laser resampling      
      jHazards();           // obstacle avoidance hint
      jVisuals();           // 
      // END: perception and actuation
      
      // BEGIN: conception
      cStone=cPos*jStone;   // transfer stones using old position (because cam img are rather delayed)
      cLocalize();          // determines position of j2b2
      mLocalize();          // determines position of marsokhod
      cMapping();           // updates map
      cRouting();           // routes the free space
      pManage();            // path planning
      mMobiles();           // determines mobility of marsokhod
      mStardusts();         // best point to follow marsokhod
      cProfile();
      cMovements();         // fills cMoves with locations of mass movements
      // END: conception
      
      // BEGIN: connection
      pRecord();            // store robots state, hash geometry if in record mode
      pTracking();          // route path in sneak/atlas
      pPlaces(pMoves,cMoves,jTog[jTOG_MOTION_DETECTED],pMOVES_SPACING);
      pPlaces(pStone,cStone,jTog[jTOG_STONED_DETECTED],pSTONE_SPACING);      
      pMovesUpon();
      pMovers();
      // END: connection
      
      while (SDL_PollEvent(&event)) {        
        
        jProcess(event);
        mProcess(event);
        cProcess(event);
        pProcess(event);

        switch (event.type) {
          case SDL_QUIT: jTic[jTOC_NEXT_SHUTDOWN]=jTic[jTOC_LAST_SENSORS]; break;
        }
      }
      
      pControl();           // strategy
      cControl();           // stopped state
      mControl();
      jControl();           // basic behavior and control of actuators

      jHUD();
      cHUD();
      pHUD();

      jDraw();              // draws cameras, front view, and map
      cBlits();             // centers the robot in the map
      pAltar();
      
      cLaserSpot.size[2]+=cLaserSpot.size[0];
      cLaserSpot.size[2]%=cLaserSpot.numel;      
      
      SDL_Flip(screen);
      SDL_FillRect(screen,NULL,SDL_MapRGBA(screen->format,255,255,255,0));
      
      if (jTog[jTOG_HEAVY_DUTY]) {
        jTog[jTOG_HEAVY_DUTY]=false;
        ownSleep_ms(5);
      } else
        ownSleep_ms(20);
    }
  }  
  cShutdown();
  mShutdown();
  jShutdown();
  jDEBUG("exit by jph");
  ownSleep_ms(50);
  return 0;
}
