// code by amaula

// to run the demo #define either
// #define SERVERCONNECTION_DEBUG
// ... or ...
// #define CONNECTION_DEBUG

#include "connectionTCP.h"

static int clientConnect(TConnectionTCP* con);
static int serverConnect(TConnectionTCP* con);

const TConnectionPtr* connectionTCPGetConnectionPtrEntry(const TConnectionTCP* con) {
  return &con->connectionPtr;
}

int connectionTCPInit(TConnectionTCP* con) {
  int r;
  con->host=NULL;
  con->port=0;
  con->hostEntry=NULL;
  con->connectionStatus=tcpStatusDisconnected;
  con->connectionStyle=tcpStyleClient;
  con->blockingType=tcpBlockingTypeBlocking;    // tcpBlockingTypeNonBlocking !!!!
  con->socket_fd=-1;
  con->data_fd=-1;
  con->connectionPtr.dataPtr=con;
  con->connectionPtr.sendFunc=(sendDataFuncPtr)connectionTCPSendData;
  con->connectionPtr.receiveFunc=(receiveDataFuncPtr)connectionTCPReceiveData;
  con->connectionPtr.receiveTimedFunc=(receiveTimedDataFuncPtr)connectionTCPReceiveDataTimed;
  con->connectionPtr.connectFunc=(connectFuncPtr)connectionTCPConnect;
  con->connectionPtr.disconnectFunc=(disconnectFuncPtr)connectionTCPDisconnect;
  con->connectionPtr.getHandleFunc=(getHandleFuncPtr)connectionTCPGetHandle;
  con->connectionPtr.deinitFunc=(deinitFuncPtr)connectionTCPDeinit;
  con->connectionPtr.getUnsentBytesFunc=(getUnsentBytesFuncPtr)connectionTCPGetUnsentBytes;
  con->connectionPtr.getPeerIPFunc=(getPeerIPFuncPtr)connectionTCPGetPeerIP;
  con->connectionPtr.getPeerPortFunc=(getPeerPortFuncPtr)connectionTCPGetPeerPort;
  // Initialize mutex. Don't want concurrent access over 'Connection'
  r=pthread_mutex_init(&con->mutex,NULL);
  assert(r == 0);               // TODO: try w/o  r
  return CONNECTIONTCP_OK;
}

int connectionTCPGetUnsentBytes(TConnectionTCP* con) {
  int value=0;
  int error=1;
#ifdef LINUX_OS
  error=ioctl(con->data_fd,SIOCOUTQ,&value);
#else
//  #warning Unsupported platform for SIOCOUTQ ioctl. connectionTCPGetUnsentBytes  will fail.
#endif
  if (error==0)
    return value;
  else
    return -1;
}

const char* connectionTCPGetPeerIP(const TConnectionTCP* con) {
  assert(con!=NULL);          // TODO: remove asserts
  return inet_ntoa(con->socket_addr.sin_addr);
}

int connectionTCPGetPeerPort(const TConnectionTCP* con) {
  assert(con!=NULL);          // TODO: remove asserts
  return (int)(con->socket_addr.sin_port);
}

int connectionTCPDeinit(TConnectionTCP* con) {
  connectionTCPDisconnect(con);
  if (con->host!=NULL) 
    free(con->host);
  /*  if (con->socket_fd != -1)
      close(con->socket_fd);*/      // TODO: try with close
  pthread_mutex_destroy(&con->mutex);  
  return 0;
}

int connectionTCPGetHandle(TConnectionTCP* con) {
  return con->data_fd;
}

int connectionTCPCreateListeningSocket(int port,struct sockaddr_in *socket_addr) {
#ifdef ENABLE_REUSEPORT
  int optionValue=0;
#endif
  int socket_fd;
  int r;
  int flag=1;
  if (port<=0||socket_addr==NULL)
    return CONNECTIONTCP_ERROR;
  // Create Socket for listening
  if ((socket_fd=socket(PF_INET,SOCK_STREAM,0))<0) {
    //dPrintG(1, "Failed to CREATE a socket!"); 
    return CONNECTIONTCP_ERROR;
  }
  // Marking the PORT REUSABLE, this will allow connections reopened before
  // the 60 seconds socket-timeout has passed.
  // NOTE; All systems don't support this. (therefore the #ifdef)
#ifdef ENABLE_REUSEPORT
  setsockopt(socket_fd,SOL_SOCKET,SO_REUSEADDR,&optionValue,sizeof(int));
#endif
  // TODO: without r
  r=setsockopt(socket_fd,IPPROTO_TCP,TCP_NODELAY,(char*)&flag,sizeof(int));
  if (r<0) {
    //dPrint(2,"Failed to set TCP_NODELAY Socket option!");
  } else {
    //dPrint(3,"Succesfully enabled TCP_NODELAY for socket %d", socket_fd);
  }  
  // Zero the address structure (required for maximum portability)
  bzero(socket_addr,sizeof(struct sockaddr_in));
  // sets few struct values
  socket_addr->sin_family=AF_INET; // Internet protocol
  socket_addr->sin_port=htons(port); // Store port
  socket_addr->sin_addr.s_addr=INADDR_ANY; // Any address  
  // Binds a name to a socket
  if (bind(socket_fd,(struct sockaddr *)socket_addr,sizeof(struct sockaddr_in))<0) {
    //dPrintG(1, "Failed to BIND a socket!");
    return CONNECTIONTCP_ERROR;
  }
  // Listens for socket connections and
  // limits the backlog of incoming connections to 20
  if (listen(socket_fd,20) != 0) {
    //dPrintG(1, "Failed to LISTEN for socket!");
    return CONNECTIONTCP_ERROR;
  }
  return socket_fd;
}

int connectionTCPSetConnectionType(TConnectionTCP* con,EConnectionTCPStyle style,EConnectionTCPBlockingType blockingType) {
  // Store blocking flag value.
  con->connectionStyle=style;
  con->blockingType=blockingType;  
  //if (style==tcpStyleClient) {
  //  con->connectionStyle = tcpStyleClient;
  //} else if (style==tcpStyleServer) {
  //  con->connectionStyle = tcpStyleServer;
  //} else {
  //  return CONNECTIONTCP_ERROR;
  //}
  return CONNECTIONTCP_OK;
}

int connectionTCPSetSocketHandle(TConnectionTCP* con,int socket_fd,struct sockaddr_in* socket_addr) {
  if (con->connectionStyle!=tcpStyleServer||socket_addr==NULL||socket_fd<0) {
    return CONNECTIONTCP_ERROR;
  }
  con->socket_fd=socket_fd;
  memcpy(&con->socket_addr,socket_addr,sizeof(struct sockaddr_in));
  return CONNECTIONTCP_OK;
}

static int serverConnect(TConnectionTCP* con) {
  int result=CONNECTIONTCP_ERROR;
  int r;
  int flag=1;
  socklen_t address_length=sizeof(con->socket_addr);
#ifdef ENABLE_REUSEPORT
  int optionValue=1;
#endif
  // Lock
  pthread_mutex_lock(&con->mutex);
  // If con->socket_fd is previously unopened, create it now.
  if (con->socket_fd==-1) {
    con->socket_fd=connectionTCPCreateListeningSocket(con->port,&con->socket_addr);
  }
  // Test whether the socket is now open
  if (con->socket_fd!=-1) {
    // Needed with accept
    do { // Accepts a new connection on a socket, waits until connection OK.
      con->data_fd=accept(con->socket_fd,(struct sockaddr*)&con->socket_addr,&address_length);
    } while (con->data_fd==-1);    
    // Only proceed if connection accepted OK.
    if (con->data_fd!=-1) {
      // Log the successfull connection.
      // dPrintG(2, "Connection from IP: %s", inet_ntoa(con->socket_addr.sin_addr));
      con->connectionStatus=tcpStatusConnected; // Assign current "connectedstate" flag
      // Set the stream NON BLOCKING! (on demand)
      if (con->blockingType==tcpBlockingTypeNonBlocking) {
	      r=fcntl(con->data_fd,F_SETFL,O_NONBLOCK);
	      assert(r==0);
      }
      // Set the socket TCP_NODELAY option
      r=setsockopt(con->data_fd,IPPROTO_TCP,TCP_NODELAY,(char *)&flag,sizeof(int));
      if (r<0) {
	      //dPrint(2,"Failed to set TCP_NODELAY Socket option for accept():ed socket %d!",con->data_fd);
      } else {                                                                                                                  
	      //dPrint(3,"Succesfully enabled TCP_NODELAY for socket %d", con->data_fd);                                                   
      }                                                                                   
      // Finally, mark the return value to be OK.
      result=CONNECTIONTCP_OK;
    }
  }
  // Unlock
  pthread_mutex_unlock(&con->mutex);
  return result;
}

static int clientConnect(TConnectionTCP *con) {
  int result=CONNECTIONTCP_ERROR;
  assert(con->host!=NULL);
  assert(con->port!=-1);
  // Lock
  pthread_mutex_lock(&con->mutex);
  // Create socket.
  if ((con->socket_fd=socket(AF_INET,SOCK_STREAM,0)) < 0) {
    //dPrintG(1, "Unable to create socket!");
  } else {
    //Zero socket addr and set required parameters
    memset(&con->socket_addr,0,sizeof(con->socket_addr));
    con->socket_addr.sin_family=AF_INET; // Set to INET type
    con->socket_addr.sin_port=htons(con->port); // Converts the port value to netform  
    // Do hostname lookup. Also ip addresses are valid.
    if ((con->hostEntry=gethostbyname(con->host))==NULL) {
      //dPrintG(1, "Invalid hostname! (%s)", hstrerror(h_errno));
      close(con->socket_fd); // clean up      
    } else {
      // Assign address if it was valid
      con->socket_addr.sin_addr=*((struct in_addr *)con->hostEntry->h_addr);
      // Now we have socket binded and stuff. Let's connect!
      if (connect(con->socket_fd,(struct sockaddr*)&con->socket_addr,sizeof(con->socket_addr))!=0) {
	  //dPrintG(1, "Error connecting to %s port %d!", 
		//con->host,
		//con->port); // TODO: why not closed?
	//close(con->socket_fd); // Do very clean! (yeah, right. ;)	
      } else {
	      memset(&con->socket_status,0,sizeof(con->socket_status)); // Zero socket status
	      con->socket_status.fd=con->socket_fd; // Set status field to current socket
	      con->socket_status.events=POLLHUP|POLLERR|POLLNVAL; // set flags
	      con->connectionStatus=tcpStatusConnected; // Assign current "connectedstate" flag	
	      con->data_fd=con->socket_fd; // Copy to data handle	
	      if (con->blockingType==tcpBlockingTypeNonBlocking) {
	        int r=fcntl(con->data_fd,F_SETFL,O_NONBLOCK);
	        assert(r==0);
	      }
	      // Finally, set result to OK
	      result = CONNECTIONTCP_OK;	
      } // if (connect...)
    } // if(con->hostEntry...)
  } // if(con->socket_fd...)
  pthread_mutex_unlock(&con->mutex);
  return result;
}

int connectionTCPConnect(TConnectionTCP* con) {
  int result=CONNECTIONTCP_ERROR;
  if (con->connectionStyle==tcpStyleClient) 
    result=clientConnect(con);  
  if (con->connectionStyle==tcpStyleServer) 
    result = serverConnect(con);  
  return result;
}

int connectionTCPDisconnect(TConnectionTCP* con) {
  int result=CONNECTIONTCP_ERROR;
  pthread_mutex_lock(&con->mutex);
  if (con->connectionStatus!=tcpStatusDisconnected) {
    // Just close the socket, effectively closes the connection
    if (con->data_fd!=-1)
      close(con->data_fd);
    // Reassign status
    con->connectionStatus=tcpStatusDisconnected;
    // Set data_fd to -1 (no longer available)
    con->data_fd=-1;
  } // if (con->connectionStatus...)  
  // UNLOCK
  pthread_mutex_unlock(&con->mutex);  
  return result;
}

int connectionTCPSetHost(TConnectionTCP* con,const char* hostname) {
  assert(hostname!=NULL);
  pthread_mutex_lock(&con->mutex);  
  if (con->host!=NULL)
    free(con->host);  
  con->host=(char*)malloc(strlen(hostname)+1);
  strcpy(con->host,hostname);
  pthread_mutex_unlock(&con->mutex);  
  return CONNECTIONTCP_OK;
}

int connectionTCPSetPort(TConnectionTCP *con, int port) {
  if (port<=0) 
    return CONNECTIONTCP_ERROR;
  con->port=port;
  return CONNECTIONTCP_OK;
}

//int connectionTCPSendDataFormatted(TConnectionTCP* con,const char* format,...) {
//  char buffer[2048];
//  int bytes;
//  va_list va; //parameter list   
//  va_start(va,format); //get list
//  bytes=vsnprintf(buffer,sizeof(buffer),format,va); //print list
//  va_end(va); //end list
//  assert(bytes<sizeof(buffer));
//  // Send the string (This is a wrapper, no own mutexes needed)
//  bytes = connectionTCPSendData(con,(unsigned char*)buffer,bytes);
//  return bytes;
//}

int connectionTCPSendData(TConnectionTCP* con,const unsigned char* buffer,int bufLen) {
  int num=0,r;
  assert(buffer!=NULL);
  assert(bufLen>=0);
  // If buflen is Zero, return zero bytes written. (which is correct)
  if (bufLen == 0) return 0;
  // LOCK
  pthread_mutex_lock(&con->mutex);
  if (con->blockingType==tcpBlockingTypeBlocking) {
    // we're free to write! This guarantees the whole buffer is sent!
    do {
      r=write(con->data_fd,buffer+num,bufLen-num);
      num+=r;
      //dPrintG(8, "Incomplete write! %d/%d bytes written", num, bufLen);
    } while (num<bufLen&&r>=0);    
    if (r<0) {
      // Despite the error checking, something went wrong.
      //dPrintG(1, "Connection error, data sending failed!");
      con->connectionStatus=tcpStatusError;
      num=-1;
    }
  } else {
    // Performing non-blocking write; therefore, writing as much
    // as possible, and returning the amount sent
    num=write(con->data_fd,buffer,bufLen);
    if (num<0) {
//      if (errno == EAGAIN) {
	// Connection is non-blocking, and BLOCKING state encountered,
	// so, returning Zero bytes written (which is the truth)
	      num = 0;      
//      } else {
//	// Otherwise, there was some error!
//	num = -1;
//      }
    }
    //dPrintG(8, "%d/%d bytes written", num, bufLen);
  }
  // UNLOCK
  pthread_mutex_unlock(&con->mutex);
  return num;
}

int connectionTCPReceiveDataTimed(TConnectionTCP* con,unsigned char* buffer,int bufLen,int timeout_ms) {
  struct pollfd p;
  int v;
  if (con->data_fd < 0)
    return -1;
  assert(bufLen>=1);
  // Assign poll() parameters
  p.fd=con->data_fd;  // File descriptor
  p.events=POLLIN;    // Only interested in poll_in
  p.revents=0;        // To satisfy valgrind http://valgrind.org/
  // Call poll(), until it succeeds ;)
  while ((v=poll(&p,1,timeout_ms))<0) ;  
  if (v==0) {
    //dPrint(10,"Poll returned Zero - Timeout! No data to read.");
  } else if (v == 1 && (p.revents & POLLIN)) {
    //dPrint(10, "Poll returned Events!");
    // LOCK
    pthread_mutex_lock(&con->mutex);    
    v=read(con->data_fd,buffer,bufLen);
    if (v<0) {
      connectionTCPDisconnect(con);
      con->connectionStatus=tcpStatusError;
      //dPrint(2, "Error reading data from Socket!");   
      v=-1;
    }
    // UNLOCK
    pthread_mutex_unlock(&con->mutex);    
  } else {
    //dPrintLE(2,"Poll returned error!");
    v = -1;
  }
  return v;
}

int connectionTCPReceiveData(TConnectionTCP* con,unsigned char* buffer,int bufLen) {
  int v;
  //printf("DATAF: %i ",con->data_fd);
  //printf("BULEN: %i ",bufLen);
  assert(con->data_fd!=-1&&bufLen>=1);
  // FIXME: HANDLE NON-BLOCKING/BLOCKING
  pthread_mutex_lock(&con->mutex);
  v=read(con->data_fd,buffer,bufLen);
  pthread_mutex_unlock(&con->mutex);
  //if (v<0) {
  //  connectionTCPDisconnect(con);
  //  con->connectionStatus=tcpStatusError;
  //  v=-1;
  //}
  //printf("DATAF: %i ",con->data_fd);
  //printf("BULEN: %i ",bufLen);
  return v;
}

int connectionTCPReceiveDataBlock(TConnectionTCP* con,unsigned char* buffer,int readBytes,int timeout) {
  int gotBytes=0,retVal;
  fd_set readfs;          // FIle descriptor set
  struct timeval maxWait; // Time to wait (still)
  if (con->data_fd<0||timeout<0)
    return -1;
  FD_ZERO(&readfs);       // ZERO descriptor set.
  FD_SET(con->data_fd, &readfs); // Assign fs
  // Assign timeout values. (Convert to timeval struct)
  // TODO: with all respect, THIS IS CRAP: !?!!!
  //maxWait.tv_sec =(timeout-(timeout%1000)) / 1000;  // Full Seconds
  //maxWait.tv_usec=(timeout%1000)*1000;              // Microseconds
  maxWait.tv_sec =timeout/1000;  // Full Seconds
  maxWait.tv_usec=timeout%1000;  // Microseconds
  do {
    if (timeout>0) {
      retVal=select(con->data_fd+1,&readfs,NULL,NULL,&maxWait); // WAIT.
      if (retVal==0) { // Timeout
        return gotBytes; // Return incomplete packet
      } else if (retVal==-1) { // error
        return -1;
      }
    }
    // No mutex needed.
    retVal=connectionTCPReceiveData(con,buffer+gotBytes,readBytes-gotBytes);
    if (retVal<0) return -1; // ERROR, read aborted
    if (retVal==0) return gotBytes;
    gotBytes+=retVal;
  } while (gotBytes<readBytes);
  return gotBytes;
}

int connectionTCPReceiveDataTextLine(TConnectionTCP* con,char* buffer,int bufLen,char endChar) {
  int gotBytes=0,retVal;
  unsigned char readChar;
  if (buffer==NULL||bufLen<1)
    return -1;
  do {
    retVal=connectionTCPReceiveDataBlock(con,&readChar,1,0); // Read one byte and sit forever for it
    if (retVal<=0)
      return -1; // Connection broken BEFORE line full. Need to quit.
    // fuck    
    *(buffer+gotBytes)=(isprint(readChar)||isspace(readChar))?readChar:'*';
    //buffer[gotBytes]
    gotBytes++;
  } while (gotBytes<bufLen&&readChar!=endChar);
  return gotBytes;
}

/// gcc -Wall connectionTCP.c -DCONNECTION_DEBUG -DLINUX_OS ../utils/owndebug.c ../utils/ownutils.c -lrt -lpthread
#ifdef CONNECTION_DEBUG

int main(void) {
  int i;
  char buffer[1024];  
  char host[100];
  int port;
  TConnectionTCP c;
  do {
    printf("Enter host and port to connect to (<host> <port>):\n");
  } while (scanf("%s %d",host,&port)!=2);
  printf("Connecting to \"%s\", port %d\n",host,port);  
  assert(connectionTCPInit(&c)==0);
  assert(connectionTCPSetHost(&c,host)==0);
  assert(connectionTCPSetPort(&c,port)==0);
  assert(connectionTCPConnect(&c)==0);
  printf("\n\nEnter data to send ('q' on a empty line to quit):\n");
  // CONNECTION IS BLOCKING!!!
  do {
    for (i=0;i<sizeof(buffer);i++) {
      fflush(stdin);
      buffer[i]=getchar();
      if (buffer[0]=='q') break;
      if (buffer[i]=='\n') {
        assert(connectionTCPSendData(&c,(unsigned char*)buffer,i+1)==i+1);
        break;
      }
    }
  } while (buffer[0]!='q');
  printf("Receiving; Send 'q' to quit\n");
  buffer[0]='f';
  do {
    int bytes;
    bytes=connectionTCPReceiveData(&c,(unsigned char*)buffer,sizeof(buffer));
    assert(bytes>0);
    buffer[bytes]='\0';
    printf("Got: \"%s\"\n",buffer);
  } while (buffer[0]!='q');
  connectionTCPDeinit(&c);
  return 0;
}
#endif //CONNECTION_DEBUG

//gcc -Wall connectionTCP.c -DSERVERCONNECTION_DEBUG -DLINUX_OS ../utils/owndebug.c ../utils/ownutils.c -lrt -lpthread
#ifdef SERVERCONNECTION_DEBUG

int main(void) {
  int c1;
  TConnectionTCP tcp;
  struct sockaddr_in sockaddr;
  int port=49401;
  int MOD=256;
  unsigned char fin[256];
  unsigned char buf[MOD];
  unsigned char msg[16];
  int head=0;
  int tail=0;
  printf("Listening on port: %d\n", port);
  printf("connectionTCPCreateListeningSocket...");
  int socket=connectionTCPCreateListeningSocket(port,&sockaddr);
  printf("ok.\n");
  printf("connectionTCPInit...");
  assert(connectionTCPInit(&tcp)==0);
  printf("ok.\n");    
  printf("connectionTCPSetPort...");
  assert(connectionTCPSetPort(&tcp,port)==0);
  printf("ok.\n");    
  printf("connectionTCPSetConnectionType...");
  assert(connectionTCPSetConnectionType(&tcp,tcpStyleServer,tcpBlockingTypeNonBlocking)==0);
  printf("ok.\n");
  printf("connectionTCPSetSocketHandle...");
  assert(connectionTCPSetSocketHandle(&tcp,socket,&sockaddr)==0);
  printf("ok.\n");  
  printf("connectionTCPConnect...WAITING...\n");
  assert(connectionTCPConnect(&tcp)==0);
  printf("ok.\n");
  printf("FID %i\n",tcp.data_fd);
  int len;
  do {    
    //printf("connectionTCPReceiveData...");
    len=connectionTCPReceiveData(&tcp,fin,sizeof(fin));
    for (c1=0;c1<len;c1++) {
			buf[head]=fin[c1];
			head++;
			head%=MOD;
    }
		len=head-tail;
		if (len<0) len+=MOD;
    if (11<=len) {
		  for (c1=0;c1<11;c1++) {
		  	msg[c1]=buf[(tail+c1)%MOD];		  	
		  	printf("%i ",msg[c1]);
		  }
		  printf("\n");
		  // clear
		  tail+=11;
		  tail%=MOD;
		  
    }
    //printf("ok. %i\n",bytes);
    //if (bytes<=0)
    //  break;
    //buffer[bytes] = '\0';
    //printf("Got: \"%s\"\n", buffer);
    //if (0<len)
    //  printf("<%i>\n",len);
  } while (1);
  printf("connectionTCPDeinit...");
  connectionTCPDeinit(&tcp);
  printf("ok.\n");
  return 0;
}

#endif //SERVERCONNECTION_DEBUG
