// code by jph

// ICR 511
// CON 0
// DOG 4

#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <linux/sockios.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#ifdef USE_RT_SCHEDULING
#include <sys/mman.h>
#include <sched.h>
#endif

#include "can4linux.h"
#include "getstat.c"
#include "connectionTCP.c"

#define         REPLY       5000  // mu.sec until chips have answered
#define         TCP         1024  // buffer for tcp recv

int main(int argc, char **argv) {

int             c1,c2,p1;
int             CID;
int             can_fid;
canmsg_t        can_msg;
Command_par_t   can_cmd;
Config_par_t    can_cfg;
CanStatusPar_t  can_sta;

unsigned char   tcp_buf[TCP];
unsigned char   tcp_rcv[TCP];
unsigned char   tcp_msg[TCP];
int             tail=0;         // ringbuffer mark
int             head=0;         // ringbuffer mark
int             MSK=0;          // bit mask that indicates which of the chips [0,1,...,8] are operational
unsigned char   NUM=0;          // number of chips that are operational

// BEGIN: init can and scan for chips
can_fid=open("/dev/can0",O_RDWR);
can_cmd.cmd=CMD_RESET;
ioctl(can_fid,CAN_IOCTL_COMMAND,&can_cmd);
can_cmd.cmd=CMD_STOP;
ioctl(can_fid,CAN_IOCTL_COMMAND,&can_cmd);  
can_cfg.target=CONF_TIMING;
can_cfg.val1=1000;              // baud rate 1000 Mbit/s
ioctl(can_fid,CAN_IOCTL_CONFIG,&can_cfg);  
can_cmd.cmd=CMD_START;
ioctl(can_fid,CAN_IOCTL_COMMAND,&can_cmd);  

ioctl(can_fid,CAN_IOCTL_STATUS,&can_sta);
for (c1=0;c1<can_sta.rx_buffer_used;c1++)
  read(can_fid,&can_msg,1);
if (can_sta.rx_buffer_used)
  printf("cleared %i messages from can buffer\n",can_sta.rx_buffer_used);

for (c1=0;c1<8;c1++) {          // all motor chips
  CID=c1<<7;
  can_msg.flags=0;
  can_msg.cob=0;
  can_msg.id=CID|2<<3;          // configuration frame
  can_msg.length=5;
  can_msg.data[0]=0x01;         // icr = 511
  can_msg.data[1]=0xff;
  can_msg.data[2]=0x00;         // pwm = 0
  can_msg.data[3]=0x00;  
  can_msg.data[4]=0x00;         // con = 0, demands a tcp communication cycle of < 0.032 s (con = 1 demands h < 0.065536 s)
  write(can_fid,&can_msg,1);
}
c1=8;                           // chip for measurements
  CID=c1<<7;
  can_msg.flags=0;
  can_msg.cob=0;
  can_msg.id=CID|2<<3;          // configuration frame
  can_msg.length=1;
  can_msg.data[0]=4;            // dog = 4
  write(can_fid,&can_msg,1);  
printf("set dead man's switch\t%6.3f sec\n",can_msg.data[0]*0.262144);

usleep(REPLY);

ioctl(can_fid,CAN_IOCTL_STATUS,&can_sta);
printf("[can]\n");
printf(" baud\t%d\n",can_sta.baud);
printf(" errors\trx %d\ttx %d\tcode %d\n",can_sta.rx_errors,can_sta.tx_errors,can_sta.error_code);
printf(" rx_buffer\t%d\t[%d]\n",can_sta.rx_buffer_used,can_sta.rx_buffer_size);
printf(" tx_buffer\t%d\t[%d]\n",can_sta.tx_buffer_used,can_sta.tx_buffer_size);

printf("[chips]\n");
NUM=can_sta.rx_buffer_used;
for (c1=0;c1<NUM;c1++) {
  read(can_fid,&can_msg,1);
  p1=can_msg.id>>7;
  printf(" %d",p1);
  MSK|=(1<<p1);
}
printf("\n");
if (9<NUM) {
  printf("abort due to message overflow!\n");
  return 0;
}
printf(" master\t\t%d #=%d\n",MSK,NUM);
if ((MSK&0xff)!=0xff)
  printf("\n---------------------------------\n\n WARNING :: SOME MOTORS INACTIVE \n\n---------------------------------\n");
// END: init can and scan for chips

// BEGIN: init tcp and wait for connection
TConnectionTCP        tcp;
struct  sockaddr_in   sockaddr;
int                   port=49401;
int                   socket;

printf("[tcp]\n server port\t%d\n", port);
socket=connectionTCPCreateListeningSocket(port,&sockaddr);
assert(connectionTCPInit(&tcp)==0);
tcp.port=port;
tcp.connectionStyle=tcpStyleServer;
tcp.blockingType=tcpBlockingTypeNonBlocking;  
tcp.socket_fd=socket;
memcpy(&tcp.socket_addr,&sockaddr,sizeof(struct sockaddr_in));
printf(" waiting for tcp connection ...\n");
assert(connectionTCPConnect(&tcp)==0);
printf(" ok.\n");
usleep(REPLY);  // requires that tcp data arrives soon
// END: init tcp and wait for connection

// BEGIN: main loop of operation
int len=0;      // number of tcp recv bytes
int cnt=0;      // ring buffer pointer
int itp=0;      // indicates: message has not completely arrived
int tmp,tnp;    // temporary
int slp=0;      // upcounts between tcp messages
int sta;        // checks consistency of can messages
int run=1;      // operation flag

while (slp<625&&run) {            // 0.5 sec until timeout exit
  len=connectionTCPReceiveData(&tcp,tcp_msg,sizeof(tcp_msg));   // poll for tcp recv
  for (c1=0;c1<len;++c1) {        // transfer tcp recv into buffer
    tcp_buf[head]=tcp_msg[c1];
    ++head;
    head%=TCP;
  }
  len=(head-tail+TCP)%TCP;        // total bytes tcp recv
  if (0<len) {
    cnt=tail;
    for (c1=0;c1<len;++c1) {      // for convenience
      tcp_rcv[c1]=tcp_buf[cnt];
      ++cnt;
      cnt%=TCP;
    }
    cnt=0;
    itp=0;
    while (cnt<len&&!itp&&run) {
    	switch (tcp_rcv[cnt]) {
    	  case 0xf0:
    	    if (cnt+33<=len) {
            for (c1=0;c1<8;c1++)        // muc motor
              if (MSK&1<<c1) {
                p1=cnt+1+c1*3;
                CID=c1<<7;
                can_msg.flags=0;
                can_msg.cob=0;
                can_msg.id=CID|0<<3;    // motor control frame
                can_msg.length=3;
                can_msg.data[0]=tcp_rcv[p1+0]; // drv|eim
                can_msg.data[1]=tcp_rcv[p1+1]; // pwm hi
                can_msg.data[2]=tcp_rcv[p1+2]; // pwm lo
                write(can_fid,&can_msg,1);
              }
            c1=8;                       // muc measurements
              if (MSK&1<<c1) {
                p1=cnt+1+c1*3;
                CID=c1<<7;
                can_msg.flags=0;
                can_msg.cob=0;
                can_msg.id=CID|0<<3;    // MID=0     
                can_msg.length=8;
                can_msg.data[0]=tcp_rcv[p1+0]; // led0|lcd 
                can_msg.data[1]=tcp_rcv[p1+1]; // led1|lcd
                can_msg.data[2]=tcp_rcv[p1+2]; // lcd
                can_msg.data[3]=tcp_rcv[p1+3]; // lcd
                can_msg.data[4]=tcp_rcv[p1+4]; // lcd
                can_msg.data[5]=tcp_rcv[p1+5]; // lcd
                can_msg.data[6]=tcp_rcv[p1+6]; // lcd
                can_msg.data[7]=tcp_rcv[p1+7]; // lcd
                write(can_fid,&can_msg,1);
              }
            cnt+=33;
            tcp_msg[0]=0xf1;
            tcp_msg[1]=MSK>>8;
            tcp_msg[2]=MSK;
            tcp_msg[3]=slp>>8;
            tcp_msg[4]=slp;                        
            usleep(REPLY); // grant chips time to response
            ioctl(can_fid,CAN_IOCTL_STATUS,&can_sta); // check if number of NUM can msg recv
            run&=can_sta.rx_buffer_used==NUM;         // abort operation if not exact number
            if (can_sta.rx_buffer_used!=NUM)  
              printf("insufficient CAN messages received\n");
            if (can_sta.rx_errors||can_sta.tx_errors) // print if there have been some errors ("errors" are NOT critical)
              printf(" errors\trx %i\ttx %i\n",can_sta.rx_errors,can_sta.tx_errors);
            for (c1=0;c1<can_sta.rx_buffer_used;c1++) {
              read(can_fid,&can_msg,1);
              p1=can_msg.id>>7;
              if (can_sta.rx_buffer_used!=NUM)
                printf(" recv\t%i\n",p1);             // show which of the chips have answered
              p1*=8;
              p1+=5;
              for (c2=0;c2<8;c2++)
                tcp_msg[p1+c2]=can_msg.data[c2];
            }
            connectionTCPSendData(&tcp,tcp_msg,77);  // 5+8*8+8
          } else 
            itp=1;
          break;
        case 0xf2:                  // configuration of chips
    	    if (cnt+41<=len) {
      	    sta=1;
            for (c1=0;c1<8;c1++) {  // checking icr,pwm for consistency
              p1=cnt+1+c1*5;
              tmp =tcp_msg[p1+0]; tmp<<=8;
              tmp+=tcp_msg[p1+1];
              sta&=255<=tmp&&tmp<=2303;
              tnp =tcp_msg[p1+2]; tnp<<=8;
              tnp+=tcp_msg[p1+3];
              sta&=tnp<=tmp;
              sta&=tcp_msg[p1+4]<20;
            }
            if (sta) {
              tmp=0;
              for (c1=0;c1<8;c1++) 
                if (MSK&1<<c1) {
                  p1=cnt+1+c1*5;
                  CID=c1<<7;
                  can_msg.flags=0;
                  can_msg.cob=0;
                  can_msg.id=CID|2<<3;  // configuration frame
                  can_msg.length=5;
                  can_msg.data[0]=tcp_msg[p1+0]; // icr hi
                  can_msg.data[1]=tcp_msg[p1+1]; // icr lo
                  can_msg.data[2]=tcp_msg[p1+2]; // pwm hi
                  can_msg.data[3]=tcp_msg[p1+3]; // pwm lo
                  can_msg.data[4]=tcp_msg[p1+4]; // con
                  write(can_fid,&can_msg,1);
                  tmp++;
                }
              tcp_msg[0]=0xf3;
              tcp_msg[1]=MSK;
              tcp_msg[2]=slp>>8;
              tcp_msg[3]=slp;
              usleep(REPLY);    // wait for reply of chips
              ioctl(can_fid,CAN_IOCTL_STATUS,&can_sta);
              run&=can_sta.rx_buffer_used==tmp;
              if (can_sta.rx_buffer_used!=tmp)
                printf("insufficient CAN messages received\n");
              if (can_sta.rx_errors||can_sta.tx_errors)
                printf(" errors\trx%i\ttx %i\n",can_sta.rx_errors,can_sta.tx_errors);
              for (c1=0;c1<can_sta.rx_buffer_used;c1++) {
                read(can_fid,&can_msg,1);
                p1=can_msg.id>>7;
                p1*=8;
                p1+=4;
                for (c2=0;c2<6;c2++)
                  tcp_msg[p1+c2]=can_msg.data[c2];
              }
              connectionTCPSendData(&tcp,tcp_msg,68); // 4+8*8
              printf(" reconfiguration of motor chips done.\n");
            } else 
              printf(" values for configuration out of bounds\n");                                    
            cnt+=41;
          } else 
            itp=1;          
          break;        
    	  default:
    	    printf(" tcp data corrupt 0x%02x\n",tcp_rcv[cnt]);
    	    cnt++;
    	    run=0;
    	    break;
  	  }
  	}
    tail+=cnt;
	  tail%=TCP;
	  slp=0;
	} else {
  	++slp;
  	usleep(800);
	}  
}
// END: main loop of operation

printf("communication terminated\n stop\t");
for (c2=0;c2<8;c2++) { // stop motors
  c1=7-c2;
  CID=c1<<7;
  can_msg.flags=0;
  can_msg.cob=0;
  can_msg.id=CID|0<<3;  // control frame
  can_msg.length=3;
  can_msg.data[0]=0x00; // drv,eim
  can_msg.data[1]=0x00; // pwm
  can_msg.data[2]=0x00; // pwm
  printf(" %i",can_msg.id>>7);
  write(can_fid,&can_msg,1);  
}
printf("\n disable dog\n");
c1=8; // top chip
  CID=c1<<7;
  can_msg.flags=0;
  can_msg.cob=0;
  can_msg.id=CID|2<<3;
  can_msg.length=1;
  can_msg.data[0]=0;    // dog
  write(can_fid,&can_msg,1);
printf("close all communication\n");
connectionTCPDeinit(&tcp);

can_cmd.cmd=CMD_STOP;
ioctl(can_fid,CAN_IOCTL_COMMAND,&can_cmd);
close(can_fid);
return 0;
}
